# pylint: disable = too-many-lines

from __future__ import annotations

import math as _math
import typing as _tp

from PyQt5.QtCore import QPointF
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QGraphicsTextItem, QUndoCommand

import trnsysGUI.massFlowSolver as _mfs
from trnsysGUI import idGenerator as _id
from trnsysGUI.CornerItem import CornerItem  # type: ignore[attr-defined]
from trnsysGUI.Node import Node  # type: ignore[attr-defined]
import trnsysGUI.PortItemBase as _pib
import trnsysGUI.SegmentItemBase as _sib


def calcDist(p1, p2):  # pylint: disable = invalid-name
    vec = p1 - p2
    norm = _math.sqrt(vec.x() ** 2 + vec.y() ** 2)
    return norm


class ConnectionBase(_mfs.MassFlowNetworkContributorMixin):
    # pylint: disable = too-many-public-methods, too-many-instance-attributes
    def __init__(self, fromPort: _pib.PortItemBase, toPort: _pib.PortItemBase, parent):
        self.logger = parent.logger

        self._fromPort = fromPort
        self._toPort = toPort
        self.displayName = ""

        self.parent = parent

        # Global
        self.id = self.parent.idGen.getID()  # pylint: disable = invalid-name
        self.connId = self.parent.idGen.getConnID()
        self.trnsysId = self.parent.idGen.getTrnsysID()

        self.segments: _tp.List[_sib.SegmentItemBase] = []  # type: ignore[name-defined]

        self.isSelected = False

        self.startNode = Node()
        self.endNode = Node()
        self.firstS: _tp.Optional[_sib.SegmentItemBase] = None  # type: ignore[name-defined]

        self.mass = 0  # comment out
        self.temperature = 0

        self.startPos = None

        self.initNew(parent)

    @property
    def fromPort(self) -> _pib.PortItemBase:
        return self._fromPort

    @property
    def toPort(self) -> _pib.PortItemBase:
        return self._toPort

    def _createSegmentItem(self, startNode, endNode):
        raise NotImplementedError()

    def isVisible(self):
        res = True
        for s in self.segments:  # pylint: disable = invalid-name
            if not s.isVisible():
                res = False

        return res

    def setMassAndTemperature(self, mass, temp):
        """
        To show the mass and temperature during mass flow visualization
        """
        self.mass = float(mass)
        self.mass = f"{self.mass:,}"

        self.temperature = temp
        for s in self.segments:  # pylint: disable = invalid-name
            replacedMass = self.mass.replace(",", "'")
            s.labelMass.setPlainText(f"M: {replacedMass} kg/h   T: {self.temperature}\u2103")

    def setDisplayName(self, newName: str) -> None:
        self.displayName = newName
        self.updateSegLabels()

    def setLabelPos(self, tup: _tp.Tuple[float, float]) -> None:
        pos = self._toPoint(tup)

        assert self.firstS

        self.firstS.label.setPos(pos)

    def setMassLabelPos(self, tup: _tp.Tuple[float, float]) -> None:
        pos = self._toPoint(tup)

        assert self.firstS

        self.firstS.labelMass.setPos(pos)

    @staticmethod
    def _toPoint(tup):
        pos = QPointF(tup[0], tup[1])
        return pos

    def setStartPort(self, newStartPort):
        self.fromPort = newStartPort
        self.startPos = newStartPort.scenePos()

    def setEndPort(self, newEndPort):
        self.toPort = newEndPort

    def setStartPos(self):
        pass

    def setEndPos(self):
        pass

    def setColor(self, value, **kwargs):
        col = QColor(0, 0, 0)

        if "mfr" in kwargs:
            if kwargs["mfr"] == "NegMfr":
                col = QColor(0, 0, 255)
            elif kwargs["mfr"] == "ZeroMfr":
                col = QColor(142, 142, 142)  # Gray
            elif kwargs["mfr"] == "min":
                col = QColor(0, 0, 204)  # deep blue
            elif kwargs["mfr"] == "max":
                col = QColor(153, 0, 0)  # deep red
            elif kwargs["mfr"] == "minTo25":
                col = QColor(0, 128, 255)  # blue
            elif kwargs["mfr"] == "25To50":
                col = QColor(102, 255, 255)  # light blue
            elif kwargs["mfr"] == "50To75":
                col = QColor(255, 153, 153)  # light red
            elif kwargs["mfr"] == "75ToMax":
                col = QColor(255, 51, 51)  # red
            else:
                col = QColor(255, 0, 0)

            for s in self.segments:  # pylint: disable = invalid-name
                self.logger.debug("Value: " + str(value))
                s.setColorAndWidthAccordingToMassflow(col, value)

        else:
            self.logger.debug("No color to set in Connection.setColor()")

    # Getters
    def getStartPoint(self):
        return QPointF(self.fromPort.scenePos())

    def getEndPoint(self):
        return QPointF(self.toPort.scenePos())

    def getNodePos(self, searchCorner):

        corners = self.getCorners()

        for i, corner in enumerate(corners):
            if corner == searchCorner:
                return i
        self.logger.debug("corner not found is " + str(searchCorner))

        return 0

    def getFirstSeg(self):
        return self.segments[0]

    def getCorners(self):
        res = []

        tempNode = self.startNode.nextN()

        while tempNode.nextN() is not None:
            res.append(tempNode.parent)
            tempNode = tempNode.nextN()

        return res

    def switchPorts(self):
        temp = self.fromPort
        self.fromPort = self.toPort
        self.toPort = temp

    # Initialization
    def initNew(self, parent):

        self.parent = parent

        self.parent.trnsysObj.append(self)

        self.displayName = "Pi" + self.fromPort.parent.displayName + "_" + self.toPort.parent.displayName

        if self.parent.editorMode == 0:
            self.logger.debug("Creating a new connection in mode 0")
            self._initSegmentM0()
        elif self.parent.editorMode == 1:
            self.logger.debug("Creating a new connection in mode 1")
            self._initializeSingleSegmentConnection()
        else:
            self.logger.debug("No valid mode during creating of connection")

        self.parent.connectionList.append(self)
        self.fromPort.connectionList.append(self)
        self.toPort.connectionList.append(self)

    def _initSegmentM0(self):
        self.startNode.setParent(self)
        self.endNode.setParent(self)

        self.startNode.setNext(self.endNode)
        self.endNode.setPrev(self.startNode)

        self.firstS = self._createSegmentItem(self.startNode, self.endNode)

        self.firstS.setLine(self.getStartPoint(), self.getEndPoint())

        self.parent.diagramScene.addItem(self.firstS)

        self.positionLabel()

    def _initializeSingleSegmentConnection(self):
        # Can be rewritten for efficiency etc. (e.g not doing initSegmentM0 and calling niceConn())

        self.startNode.setParent(self)
        self.endNode.setParent(self)

        self.startNode.setNext(self.endNode)
        self.endNode.setPrev(self.startNode)

        self.firstS = self._createSegmentItem(self.startNode, self.endNode)

        self.firstS.setLine(self.getStartPoint(), self.getEndPoint())

        self.parent.diagramScene.addItem(self.firstS)

        self.niceConn()

        self.positionLabel()

    def loadSegments(self, segmentsCorners):
        self.clearConn()

        tempNode = self.startNode

        rad = self.getRadius()

        for segmentsCorner in segmentsCorners:
            cor = CornerItem(-rad, -rad, 2 * rad, 2 * rad, tempNode, tempNode.nextN(), self)

            cor.setPos(float(segmentsCorner[0]), float(segmentsCorner[1]))
            cor.setFlag(cor.ItemSendsScenePositionChanges, True)

            cor.setZValue(100)
            cor.setVisible(True)

            self.parent.diagramScene.addItem(cor)

            cor.node.nextNode.setPrev(cor.node)
            cor.node.prevNode.setNext(cor.node)
            self._createSegmentItem(tempNode, cor.node)

            tempNode = cor.node

            self.printConn()

        self._createSegmentItem(tempNode, tempNode.nextN())

        self.printConn()

        for s in self.segments:  # pylint: disable = invalid-name
            pos1 = None
            pos2 = None

            if isinstance(s.startNode.parent, ConnectionBase) and s.startNode.prevNode is None:
                pos1 = s.startNode.parent.fromPort.scenePos().x(), s.startNode.parent.fromPort.scenePos().y()
            if isinstance(s.endNode.parent, ConnectionBase) and s.endNode.nextNode is None:
                pos2 = s.endNode.parent.toPort.scenePos().x(), s.endNode.parent.toPort.scenePos().y()

            if isinstance(s.startNode.parent, CornerItem):
                self.logger.debug(
                    str(s.startNode.parent) + " " + str(s.startNode) + "cor " + str(s.startNode.parent.scenePos())
                )
                pos1 = s.startNode.parent.scenePos().x(), s.startNode.parent.scenePos().y()
            if isinstance(s.endNode.parent, CornerItem):
                pos2 = s.endNode.parent.scenePos().x(), s.endNode.parent.scenePos().y()

            self.logger.debug("pos1 is " + str(pos1))
            self.logger.debug("pos2 is " + str(pos2))
            s.setLine(pos1[0], pos1[1], pos2[0], pos2[1])

            self.parent.diagramScene.addItem(s)

        self.firstS = self.segments[0]

        self.positionLabel()

    # Label related
    def setLabelVisible(self, isVisible: bool) -> None:
        assert self.firstS
        self.firstS.setLabelVisible(isVisible)

    def toggleLabelVisible(self) -> None:
        assert self.firstS
        self.firstS.toggleLabelVisible()

    def updateSegLabels(self):
        for s in self.segments:  # pylint: disable = invalid-name
            s.label.setPlainText(self.displayName)

    def positionLabel(self):
        self.firstS.label.setPos(self.getStartPoint())
        self.firstS.labelMass.setPos(self.getStartPoint())
        self.rotateLabel()

    def rotateLabel(self):
        angle = 0 if self.firstS.isHorizontal() else 90
        self.firstS.label.setRotation(angle)

    def setMassFlowLabelVisible(self, isVisible: bool) -> None:
        assert self.firstS
        self.firstS.setMassFlowLabelVisible(isVisible)

    def toggleMassFlowLabelVisible(self) -> None:
        assert self.firstS
        self.firstS.toggleMassFlowLabelVisible()

    def getRadius(self):
        raise NotImplementedError()

    # Makes 90deg angles of connection
    def niceConn(self):  # pylint: disable = too-many-locals, too-many-statements
        """
        Creates the segments and corners depending on the side of the fromPort and toPort
        Returns
        -------

        """
        # Here different cases can be implemented using self.PORT.side as sketched on paper
        rad = self.getRadius()

        self.logger.debug(
            "FPort " + str(self.fromPort) + " has side " + str(self.fromPort.side) + " has " + str(self.fromPort.name)
        )
        self.logger.debug(
            "FPort " + str(self.fromPort) + " has side " + str(self.fromPort.side) + " has " + str(self.fromPort.name)
        )

        if (self.fromPort.side == 2) and (self.toPort.side == 2):
            self.fromPort.createdAtSide = 2
            self.toPort.createdAtSide = 2
            self.logger.debug("NiceConn 2 to 2")
            portOffset = 30
            self.clearConn()

            corner1 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, self.startNode, None, self)
            corner2 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, corner1.node, None, self)
            corner3 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, corner2.node, None, self)
            corner4 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, corner3.node, self.endNode, self)

            corner1.node.setNext(corner2.node)
            corner2.node.setNext(corner3.node)
            corner3.node.setNext(corner4.node)

            seg1 = self._createSegmentItem(self.startNode, corner1.node)
            seg2 = self._createSegmentItem(corner1.node, corner2.node)
            seg3 = self._createSegmentItem(corner2.node, corner3.node)
            seg4 = self._createSegmentItem(corner3.node, corner4.node)
            seg5 = self._createSegmentItem(corner4.node, self.endNode)

            self.startNode.setNext(corner1.node)
            self.endNode.setPrev(corner4.node)

            self.parent.diagramScene.addItem(seg1)
            self.parent.diagramScene.addItem(seg2)
            self.parent.diagramScene.addItem(seg3)
            self.parent.diagramScene.addItem(seg4)
            self.parent.diagramScene.addItem(seg5)

            self.parent.diagramScene.addItem(corner1)
            self.parent.diagramScene.addItem(corner2)
            self.parent.diagramScene.addItem(corner3)
            self.parent.diagramScene.addItem(corner4)

            pos1 = self.fromPort.scenePos()
            pos2 = self.toPort.scenePos()

            baseLineHeight = max(pos1.y(), pos2.y()) + 100.6

            p1 = QPointF(pos1.x() + portOffset, pos1.y())  # pylint: disable = invalid-name
            p2 = QPointF(p1.x(), baseLineHeight)  # pylint: disable = invalid-name
            p3 = QPointF(pos2.x() + portOffset, baseLineHeight)  # pylint: disable = invalid-name
            p4 = QPointF(p3.x(), pos2.y())  # pylint: disable = invalid-name

            seg1.setLine(pos1, p1)
            seg2.setLine(p1, p2)
            seg3.setLine(p2, p3)
            seg4.setLine(p3, p4)
            seg5.setLine(p4, pos2)

            corner1.setFlag(corner1.ItemSendsScenePositionChanges, True)
            corner2.setFlag(corner2.ItemSendsScenePositionChanges, True)
            corner3.setFlag(corner3.ItemSendsScenePositionChanges, True)
            corner4.setFlag(corner4.ItemSendsScenePositionChanges, True)

            corner1.setZValue(100)
            corner2.setZValue(100)
            corner3.setZValue(100)
            corner4.setZValue(100)

            corner1.setPos(p1)
            corner2.setPos(p2)
            corner3.setPos(p3)
            corner4.setPos(p4)

        elif (self.fromPort.side == 0) and (self.toPort.side == 0):
            self.fromPort.createdAtSide = 0
            self.toPort.createdAtSide = 0
            self.logger.debug("NiceConn 0 to 0")
            portOffset = 30
            self.clearConn()

            corner1 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, self.startNode, None, self)
            corner2 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, corner1.node, None, self)
            corner3 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, corner2.node, None, self)
            corner4 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, corner3.node, self.endNode, self)

            corner1.node.setNext(corner2.node)
            corner2.node.setNext(corner3.node)
            corner3.node.setNext(corner4.node)

            seg1 = self._createSegmentItem(self.startNode, corner1.node)
            seg2 = self._createSegmentItem(corner1.node, corner2.node)
            seg3 = self._createSegmentItem(corner2.node, corner3.node)
            seg4 = self._createSegmentItem(corner3.node, corner4.node)
            seg5 = self._createSegmentItem(corner4.node, self.endNode)

            self.startNode.setNext(corner1.node)
            self.endNode.setPrev(corner4.node)

            self.parent.diagramScene.addItem(seg1)
            self.parent.diagramScene.addItem(seg2)
            self.parent.diagramScene.addItem(seg3)
            self.parent.diagramScene.addItem(seg4)
            self.parent.diagramScene.addItem(seg5)

            self.parent.diagramScene.addItem(corner1)
            self.parent.diagramScene.addItem(corner2)
            self.parent.diagramScene.addItem(corner3)
            self.parent.diagramScene.addItem(corner4)

            pos1 = self.fromPort.scenePos()
            pos2 = self.toPort.scenePos()

            baseLineHeight = max(pos1.y(), pos2.y()) + 100.6

            p1 = QPointF(pos1.x() - portOffset, pos1.y())  # pylint: disable = invalid-name
            p2 = QPointF(p1.x(), baseLineHeight)  # pylint: disable = invalid-name
            p3 = QPointF(pos2.x() - portOffset, baseLineHeight)  # pylint: disable = invalid-name
            p4 = QPointF(p3.x(), pos2.y())  # pylint: disable = invalid-name

            seg1.setLine(pos1, p1)
            seg2.setLine(p1, p2)
            seg3.setLine(p2, p3)
            seg4.setLine(p3, p4)
            seg5.setLine(p4, pos2)

            corner1.setFlag(corner1.ItemSendsScenePositionChanges, True)
            corner2.setFlag(corner2.ItemSendsScenePositionChanges, True)
            corner3.setFlag(corner3.ItemSendsScenePositionChanges, True)
            corner4.setFlag(corner4.ItemSendsScenePositionChanges, True)

            corner1.setZValue(100)
            corner2.setZValue(100)
            corner3.setZValue(100)
            corner4.setZValue(100)

            corner1.setPos(p1)
            corner2.setPos(p2)
            corner3.setPos(p3)
            corner4.setPos(p4)

        elif self.fromPort.side == 1:
            self.fromPort.createdAtSide = 1
            # pylint: disable = fixme
            # todo :  when rotated, it cause a problem because side gets changed

            self.logger.debug("NiceConn from 1")
            portOffset = 30
            self.clearConn()

            pos1 = self.fromPort.scenePos()
            pos2 = self.toPort.scenePos()

            if pos2.y() <= pos1.y():
                corner1 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, self.startNode, self.endNode, self)

                seg1 = self._createSegmentItem(self.startNode, corner1.node)
                seg2 = self._createSegmentItem(corner1.node, self.endNode)

                self.startNode.setNext(corner1.node)
                self.endNode.setPrev(corner1.node)

                self.parent.diagramScene.addItem(seg1)
                self.parent.diagramScene.addItem(seg2)
                self.parent.diagramScene.addItem(corner1)

                # position of the connecting node
                p1 = QPointF(pos1.x(), pos2.y() - 0.333)  # pylint: disable = invalid-name

                seg1.setLine(pos1, p1)
                seg2.setLine(p1, pos2)

                corner1.setFlag(corner1.ItemSendsScenePositionChanges, True)
                corner1.setZValue(100)
                self.fromPort.setZValue(100)
                self.toPort.setZValue(100)

                corner1.setPos(p1)
                self.firstS = self.getFirstSeg()

            else:
                self.logger.debug("To port below from port")
                corner1 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, self.startNode, None, self)
                corner2 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, corner1.node, self.endNode, self)

                corner1.node.setNext(corner2.node)

                seg1 = self._createSegmentItem(self.startNode, corner1.node)
                seg2 = self._createSegmentItem(corner1.node, corner2.node)
                seg3 = self._createSegmentItem(corner2.node, self.endNode)

                self.startNode.setNext(corner1.node)
                self.endNode.setPrev(corner2.node)

                self.parent.diagramScene.addItem(seg1)
                self.parent.diagramScene.addItem(seg2)
                self.parent.diagramScene.addItem(seg3)

                self.parent.diagramScene.addItem(corner1)
                self.parent.diagramScene.addItem(corner2)

                offsetPoint = pos1.y() - 15.666

                help_point_1 = QPointF(pos1.x(), offsetPoint)  # pylint: disable = invalid-name
                help_point_2 = QPointF(pos2.x(), offsetPoint)  # pylint: disable = invalid-name

                seg1.setLine(pos1, help_point_1)
                seg2.setLine(help_point_1, help_point_2)
                seg3.setLine(help_point_2, pos2)

                self.printConnNodes()
                corner1.setFlag(corner1.ItemSendsScenePositionChanges, True)
                corner2.setFlag(corner2.ItemSendsScenePositionChanges, True)

                corner1.setZValue(100)
                corner2.setZValue(100)
                self.fromPort.setZValue(100)
                self.toPort.setZValue(100)

                corner1.setPos(help_point_1)
                corner2.setPos(help_point_2)
                self.firstS = self.getFirstSeg()

        elif self.fromPort.side == 3:
            self.fromPort.createdAtSide = 3

            self.logger.debug("NiceConn from 1")
            portOffset = 30
            self.clearConn()

            pos1 = self.fromPort.scenePos()
            pos2 = self.toPort.scenePos()

            if pos2.y() >= pos1.y():
                corner1 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, self.startNode, self.endNode, self)

                seg1 = self._createSegmentItem(self.startNode, corner1.node)
                seg2 = self._createSegmentItem(corner1.node, self.endNode)

                self.startNode.setNext(corner1.node)
                self.endNode.setPrev(corner1.node)

                self.parent.diagramScene.addItem(seg1)
                self.parent.diagramScene.addItem(seg2)
                self.parent.diagramScene.addItem(corner1)

                # position of the connecting node
                p1 = QPointF(pos1.x(), pos2.y() - 0.333)  # pylint: disable = invalid-name

                seg1.setLine(pos1, p1)
                seg2.setLine(p1, pos2)

                corner1.setFlag(corner1.ItemSendsScenePositionChanges, True)
                corner1.setZValue(100)
                self.fromPort.setZValue(100)
                self.toPort.setZValue(100)

                corner1.setPos(p1)
                self.firstS = self.getFirstSeg()
            else:
                self.logger.debug("To port above from port")
                corner1 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, self.startNode, None, self)
                corner2 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, corner1.node, self.endNode, self)

                corner1.node.setNext(corner2.node)

                seg1 = self._createSegmentItem(self.startNode, corner1.node)
                seg2 = self._createSegmentItem(corner1.node, corner2.node)
                seg3 = self._createSegmentItem(corner2.node, self.endNode)

                self.startNode.setNext(corner1.node)
                self.endNode.setPrev(corner2.node)

                self.parent.diagramScene.addItem(seg1)
                self.parent.diagramScene.addItem(seg2)
                self.parent.diagramScene.addItem(seg3)

                self.parent.diagramScene.addItem(corner1)
                self.parent.diagramScene.addItem(corner2)

                offsetPoint = pos1.y() + 15.666

                help_point_1 = QPointF(pos1.x(), offsetPoint)  # pylint: disable = invalid-name
                help_point_2 = QPointF(pos2.x(), offsetPoint)  # pylint: disable = invalid-name

                seg1.setLine(pos1, help_point_1)
                seg2.setLine(help_point_1, help_point_2)
                seg3.setLine(help_point_2, pos2)

                self.printConnNodes()

                corner1.setFlag(corner1.ItemSendsScenePositionChanges, True)
                corner2.setFlag(corner2.ItemSendsScenePositionChanges, True)

                corner1.setZValue(100)
                corner2.setZValue(100)
                self.fromPort.setZValue(100)
                self.toPort.setZValue(100)
                self.logger.debug("Here in niceconn")

                corner1.setPos(help_point_1)
                corner2.setPos(help_point_2)
                self.firstS = self.getFirstSeg()

        else:
            pos1 = self.fromPort.scenePos()
            pos2 = self.toPort.scenePos()

            self.fromPort.createdAtSide = self.fromPort.side
            self.toPort.createdAtSide = self.toPort.side
            self.logger.debug("Ports are directed to each other")
            self.clearConn()

            corner1 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, self.startNode, None, self)
            corner2 = CornerItem(-rad, -rad, 2 * rad, 2 * rad, corner1.node, self.endNode, self)

            corner1.node.setNext(corner2.node)

            seg1 = self._createSegmentItem(self.startNode, corner1.node)
            seg2 = self._createSegmentItem(corner1.node, corner2.node)
            seg3 = self._createSegmentItem(corner2.node, self.endNode)

            self.startNode.setNext(corner1.node)
            self.endNode.setPrev(corner2.node)

            self.parent.diagramScene.addItem(seg1)
            self.parent.diagramScene.addItem(seg2)
            self.parent.diagramScene.addItem(seg3)

            self.parent.diagramScene.addItem(corner1)
            self.parent.diagramScene.addItem(corner2)

            midx = pos1.x() + 0.5 * (pos2.x() - pos1.x())

            seg1.setLine(pos1.x(), pos1.y(), midx, pos1.y())
            seg2.setLine(midx, pos1.y(), midx, pos2.y())
            seg3.setLine(midx, pos2.y(), pos2.x(), pos2.y())
            self.printConnNodes()

            corner1.setFlag(corner1.ItemSendsScenePositionChanges, True)
            corner2.setFlag(corner2.ItemSendsScenePositionChanges, True)

            corner1.setZValue(100)
            corner2.setZValue(100)
            self.toPort.setZValue(100)
            self.fromPort.setZValue(100)

            self.logger.debug("Here in niceconn")

            help_point_1 = QPointF(midx, pos1.y())  # pylint: disable = invalid-name
            help_point_2 = QPointF(midx, pos2.y())  # pylint: disable = invalid-name

            corner1.setPos(help_point_1)
            corner2.setPos(help_point_2)
            self.firstS = self.getFirstSeg()

            self.logger.debug("Conn has now " + str(self.firstS))

    # Unused
    def buildBridges(self):  # pylint: disable = too-many-locals, too-many-statements
        # This function finds the colliding line segments and creates the interrupted line effect

        for s in self.segments:  # pylint: disable = invalid-name
            col = s.collidingItems()
            # Why do the child segments have again colliding Items?

            for c in col:  # pylint: disable = invalid-name
                if isinstance(c, _sib.SegmentItemBase):
                    # Both have no bridge and do not collide at the endpoints:
                    if (c.endNode is not s.startNode) and (c.startNode is not s.endNode):
                        qp1 = s.line().p1()
                        qp2 = s.line().p2()
                        qp1_ = c.line().p1()  # pylint: disable = invalid-name
                        qp2_ = c.line().p2()  # pylint: disable = invalid-name

                        eps2 = 5
                        if abs(qp1.x() - qp2.x()) < eps2 or abs(qp1_.x() - qp2_.x()) < eps2:
                            self.logger.debug("Can't build bridge because one segment is almost verical")

                        else:
                            node1 = Node()
                            node2 = Node()

                            node1.setPrev(s.startNode)
                            node1.setNext(node2)
                            node2.setPrev(node1)
                            node2.setNext(s.endNode)
                            node1.setParent(self)
                            node2.setParent(self)

                            s.startNode.setNext(node1)
                            s.endNode.setPrev(node2)

                            s.firstChild = self._createSegmentItem(s.startNode, node1)
                            s.secondChild = self._createSegmentItem(node2, s.endNode)

                            s.hasBridge = True
                            c.hasBridge = True
                            s.bridgedSegment = c
                            c.bridgedSegment = s

                            s.firstChild.setVisible(False)
                            s.secondChild.setVisible(False)

                            self.parent.diagramScene.addItem(s.firstChild)
                            self.parent.diagramScene.addItem(s.secondChild)

                            # pylint: disable = invalid-name
                            n1 = (qp1.y() * qp2.x() - qp2.y() * qp1.x()) / (qp2.x() - qp1.x())
                            # pylint: disable = invalid-name
                            n2 = (qp1_.y() * qp2_.x() - qp2_.y() * qp1_.x()) / (qp2_.x() - qp1_.x())

                            m1 = (qp2.y() - qp1.y()) / (qp2.x() - qp1.x())  # pylint: disable = invalid-name
                            m2 = (qp2_.y() - qp1_.y()) / (qp2_.x() - qp1_.x())  # pylint: disable = invalid-name

                            collisionPos = QPointF((n1 - n2) / (m2 - m1), m1 * (n1 - n2) / (m2 - m1) + n1)

                            vecp2p1 = qp2 - qp1  # pylint: disable = invalid-name
                            vecp2p1_ = qp2_ - qp1_  # pylint: disable = invalid-name

                            normVec = _math.sqrt(vecp2p1.x() ** 2 + vecp2p1.y() ** 2)
                            normVec_ = _math.sqrt(vecp2p1_.x() ** 2 + vecp2p1_.y() ** 2)  # pylint:disable=invalid-name

                            # Compute angle between lines
                            scalarProd = vecp2p1.x() * vecp2p1_.x() + vecp2p1.y() + vecp2p1_.y()
                            angleBetween = _math.acos(scalarProd / (normVec * normVec_))

                            # Almost working: if line approaches 90 deg or if lines touch because too steep, problem
                            f = 10  # pylint: disable = invalid-name
                            distFactor = max(f * 1 / angleBetween, f)

                            normVecp2p1 = QPointF(distFactor / normVec * vecp2p1.x(),
                                                  distFactor / normVec * vecp2p1.y())

                            newPos1 = collisionPos - normVecp2p1
                            newPos2 = collisionPos + normVecp2p1

                            s.firstChild.setLine(qp1.x(), qp1.y(), newPos1.x(), newPos1.y())
                            s.secondChild.setLine(newPos2.x(), newPos2.y(), qp2.x(), qp2.y())

                            s.firstChild.setVisible(True)
                            s.secondChild.setVisible(True)

                            s.hide()

                            self.parent.diagramScene.removeItem(s)
                            self.segments.remove(s)

    # Delete a connection
    def clearConn(self):
        # Deletes all segments and corners in connection
        walker = self.endNode

        items = self.parent.diagramScene.items()

        for item in items:
            if isinstance(item, _sib.SegmentItemBase):
                if item.startNode.lastNode() is self.endNode or item.startNode.firstNode() is self.startNode:
                    self.parent.diagramScene.removeItem(item)
            if isinstance(item, QGraphicsTextItem):
                if isinstance(item.parent, _pib.PortItemBase):
                    self.logger.debug("it has " + str(item.parent()))
                    self.logger.debug("Deleting it")
                    self.parent.diagramScene.removeItem(item)

        for seg in self.segments:
            self.parent.diagramScene.removeItem(seg.label)

        self.segments.clear()

        if walker.prevN() is not self.startNode:
            walker = walker.prevN()

            while walker.prevN() is not self.startNode:
                if not isinstance(walker.parent, ConnectionBase):
                    self.parent.diagramScene.removeItem(walker.parent)
                else:
                    self.logger.debug("Caution, this is a disrupt.")
                walker = walker.prevN()

            if not isinstance(walker.parent, ConnectionBase):
                self.parent.diagramScene.removeItem(walker.parent)
            else:
                self.logger.debug("Caution.")

            self.startNode.setNext(self.endNode)
            self.endNode.setPrev(self.startNode)

    def deleteConn(self):
        self.clearConn()
        self.fromPort.connectionList.remove(self)
        self.toPort.connectionList.remove(self)

        if self not in self.parent.trnsysObj:
            return

        self.parent.trnsysObj.remove(self)
        self.parent.connectionList.remove(self)

    def createDeleteUndoCommandAndAddToStack(self) -> None:
        deleteConnectionCommand = self.createDeleteUndoCommand()
        self.parent.parent().undoStack.push(deleteConnectionCommand)

    def createDeleteUndoCommand(self, parentCommand: _tp.Optional[QUndoCommand] = None) -> QUndoCommand:
        raise NotImplementedError()

    # Gradient related
    def totalLength(self):
        summedLength = 0

        for segment in self.segments:
            summedLength += calcDist(segment.line().p1(), segment.line().p2())
        return summedLength

    def partialLength(self, node):
        # Returns the cummulative length of line up to given node
        # Assumes that segments is ordered correctly!
        res = 0
        if node == self.startNode:
            return res

        for i in self.segments:
            res += calcDist(i.line().p1(), i.line().p2())
            if i.endNode == node:
                break

        return res

    def updateSegGrads(self):
        for s in self.segments:  # pylint: disable = invalid-name
            s.updateGrad()

    # Invert connection
    def invertConnection(self):
        # Invert segment list
        self.segments.reverse()

        # Invert nodes
        self.invertNodes()

        # Invert ports
        temp = self.toPort
        self.toPort = self.fromPort
        self.fromPort = temp

        startingNode = self.startNode
        self.startNode = self.endNode
        self.endNode = startingNode

        for s in self.segments:  # pylint: disable = invalid-name
            temp2 = s.startNode
            s.startNode = s.endNode
            s.endNode = temp2

            temp3 = s.firstChild
            s.firstChild = s.secondChild
            s.secondChild = temp3

            s.setLine(s.line().p2().x(), s.line().p2().y(), s.line().p1().x(), s.line().p1().y())

            s.updateGrad()

    def invertNodes(self):
        temp = None
        element = self.startNode

        while element.nextN() is not None:
            temp = element.nextN()
            previousNode = element.prevN()
            nextNode = element.nextN()

            element.setNext(previousNode)
            element.setPrev(nextNode)
            element = temp

        previousNode = element.prevN()
        nextNode = element.nextN()
        element.setNext(previousNode)
        element.setPrev(nextNode)

    def selectConnection(self):
        for s in self.segments:  # pylint: disable = invalid-name
            s.setSelect(True)

        self.isSelected = True

        self.setLabelsSelected(True)

    def deselectConnection(self):
        for s in self.segments:  # pylint: disable = invalid-name
            s.updateGrad()

        self.isSelected = False

        self.setLabelsSelected(False)

    def setLabelsSelected(self, isSelected: bool) -> None:
        assert self.firstS
        self._setBold(self.firstS.label, isSelected)
        self._setBold(self.firstS.labelMass, isSelected)

    @staticmethod
    def _setBold(label: QGraphicsTextItem, isBold: bool) -> None:
        originalFontCopy = label.font()
        originalFontCopy.setBold(isBold)
        label.setFont(originalFontCopy)

    # Debug
    def inspectConn(self):
        self.parent.listV.clear()
        self.parent.listV.addItem("Display name: " + self.displayName)
        self.parent.listV.addItem("Parent: " + str(self.parent))
        self.parent.listV.addItem("fromPort: " + str(self.fromPort) + str(self.fromPort.id))
        self.parent.listV.addItem("toPort: " + str(self.toPort) + str(self.toPort.id))

    def printConn(self):
        self.logger.debug("------------------------------------")
        self.logger.debug("This is the printout of connection: " + self.displayName + str(self))
        self.logger.debug("It has fromPort " + str(self.fromPort))
        self.logger.debug("It has toPort " + str(self.toPort))
        self.logger.debug("It has startNode " + str(self.startNode))
        self.logger.debug("It has endNode " + str(self.endNode))

        self.logger.debug("\n")
        self.logger.debug("It goes from " + self.fromPort.parent.displayName + " to " + self.toPort.parent.displayName)
        self.logger.debug("------------------------------------")

    def printConnNodes(self):
        self.logger.debug("These are the nodes: ")

        element = self.startNode
        while element.nextN() is not None:
            self.logger.debug(
                "Node is "
                + str(element)
                + " has nextNode "
                + str(element.nextN())
                + " has pL "
                + str(self.partialLength(element))
            )
            element = element.nextN()

        self.logger.debug("Node is " + str(element) + " has nextNode " + str(element.nextN()))

    def printConnSegs(self):
        self.logger.debug("These are the segments in order")
        for s in self.segments:  # pylint: disable = invalid-name
            self.logger.debug(
                "Segment is "
                + str(s)
                + " has global id "
                + str(s.id)
                + " has startnode "
                + str(s.startNode)
                + " endnode "
                + str(s.endNode)
            )

    # Saving / Loading
    def encode(self):
        raise NotImplementedError()

    def decode(self, i):
        raise NotImplementedError()

    # Export related
    def exportBlackBox(self):  # pylint: disable = no-self-use
        return "noBlackBoxOutput", []

    def exportPumpOutlets(self):  # pylint: disable = no-self-use
        return "", 0

    def exportMassFlows(self):  # pylint: disable = no-self-use
        return "", 0

    def exportDivSetting1(self):  # pylint: disable = no-self-use
        return "", 0

    def exportDivSetting2(self, nUnit):  # pylint: disable = no-self-use
        return "", nUnit

    def getInternalPiping(self) -> _mfs.InternalPiping:
        raise NotImplementedError()

    def exportPipeAndTeeTypesForTemp(self, startingUnit):
        raise NotImplementedError()

    def findStoragePort(self, virtualBlock):
        portToPrint = None
        for port in virtualBlock.inputs + virtualBlock.outputs:
            if self in port.connectionList:
                # Found the port of the generated block adjacent to this pipe
                # Assumes 1st connection is with storageTank
                if self.fromPort == port:
                    if self.toPort.connectionList[0].fromPort == self.toPort:
                        portToPrint = self.toPort.connectionList[0].toPort
                    else:
                        portToPrint = self.toPort.connectionList[0].fromPort
                else:
                    if self.fromPort.connectionList[0].fromPort == self.fromPort:
                        portToPrint = self.fromPort.connectionList[0].toPort
                    else:
                        portToPrint = self.fromPort.connectionList[0].fromPort
        return portToPrint

    def assignIDsToUninitializedValuesAfterJsonFormatMigration(self, generator: _id.IdGenerator) -> None:
        pass



# pylint: skip-file
# type: ignore

from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import (
    QDialog,
    QLabel,
    QLineEdit,
    QPushButton,
    QCheckBox,
    QHBoxLayout,
    QGridLayout,
    QTabWidget,
    QWidget,
    QMessageBox,
)

import trnsysGUI.pump as _pmp


class PumpDlg(QDialog):
    def __init__(self, pump: _pmp.Pump, parent=None):
        super(PumpDlg, self).__init__(parent)
        nameLabel = QLabel("Name:")
        self.block = pump
        self.le = QLineEdit(self.block.label.toPlainText())
        self.setWindowIcon(QIcon(pump.pixmap()))
        self.okButton = QPushButton("OK")
        self.cancelButton = QPushButton("Cancel")

        self.hFlipBoxLabel = QLabel("Horizontal Flip")
        self.vFlipBoxLabel = QLabel("Vertical Flip")
        self.hFlipBox = QCheckBox()
        self.vFlipBox = QCheckBox()

        self.vFlipBox.setTristate(False)
        self.hFlipBox.setTristate(False)
        self.hFlipBox.setCheckState(self.block.flippedH * 2)
        self.vFlipBox.setCheckState(self.block.flippedV * 2)

        flipLayout = QHBoxLayout()
        flipLayout.addWidget(self.hFlipBoxLabel)
        flipLayout.addWidget(self.hFlipBox)
        flipLayout.addWidget(self.vFlipBoxLabel)
        flipLayout.addWidget(self.vFlipBox)

        buttonLayout = QHBoxLayout()
        buttonLayout.addStretch()
        buttonLayout.addWidget(self.okButton)
        buttonLayout.addWidget(self.cancelButton)
        tab1Layout = QGridLayout(self)
        tab1Layout.addWidget(nameLabel, 0, 0)
        tab1Layout.addWidget(self.le, 0, 1)
        tab1Layout.addLayout(flipLayout, 1, 0, 3, 0)

        positionLayout = QHBoxLayout()
        self.PumpPowerLabel = QLabel("Mass Flow Rate")
        self.LineEdit = QLineEdit(str(self.block.massFlowRateInKgPerH))
        self.PumpPowerLabel2 = QLabel("kg/h")
        positionLayout.addWidget(self.PumpPowerLabel)
        positionLayout.addWidget(self.LineEdit)
        positionLayout.addWidget(self.PumpPowerLabel2)

        # Initialize tab screen
        self.tabs = QTabWidget()
        self.tab1 = QWidget()
        self.tab2 = QWidget()

        # Add tabs
        self.tabs.addTab(self.tab1, "Diagram")
        self.tabs.addTab(self.tab2, "Adjust Mass Flow Rate")
        self.tab1.setLayout(tab1Layout)
        self.tab2.setLayout(positionLayout)

        self.layout2 = QGridLayout(self)
        self.layout2.addWidget(self.tabs, 0, 0)
        self.layout2.addLayout(buttonLayout, 2, 0, 3, 0)
        self.setLayout(self.layout2)

        self.okButton.clicked.connect(self.acceptedEdit)
        self.cancelButton.clicked.connect(self.cancel)
        self.hFlipBox.stateChanged.connect(self.setNewFlipStateH)
        self.vFlipBox.stateChanged.connect(self.setNewFlipStateV)
        self.setWindowTitle("Properties")
        self.show()

    def acceptedEdit(self):
        print("Changing displayName")
        newName = self.le.text()
        self.setPumpPower()
        if newName.lower() == str(self.block.displayName).lower():
            self.close()
        elif newName == "":
            msgb = QMessageBox()
            msgb.setText("Please Enter a name!")
            msgb.exec()
        elif self.parent().nameExists(newName) or self.parent().nameExistsInDdckFolder(newName):
            msgb = QMessageBox()
            msgb.setText("Name already exist!")
            msgb.exec()
        else:
            self.block.label.setPlainText(newName)
            self.block.displayName = newName
            self.close()

    def setNewFlipStateH(self, state):
        self.block.updateFlipStateH(state)
        self.block.updateSidesFlippedH()

    def setNewFlipStateV(self, state):
        self.block.updateFlipStateV(state)
        self.block.updateSidesFlippedV()

    def setPumpPower(self):
        self.block.massFlowRateInKgPerH = int(self.LineEdit.text())

    def cancel(self):
        self.close()



from __future__ import annotations

import typing as _tp

import PyQt5.QtWidgets as _qtw

import trnsysGUI.connection.doublePipeConnection as _dpc
import trnsysGUI.doublePipePortItem as _dpi

if _tp.TYPE_CHECKING:
    import trnsysGUI.diagram.Editor as _ed


class CreateDoublePipeConnectionCommand(_qtw.QUndoCommand):
    def __init__(
            self,
            fromPort: _dpi.DoublePipePortItem,
            toPort: _dpi.DoublePipePortItem,
            editor: _ed.Editor,  # type: ignore[name-defined]
    ) -> None:
        super().__init__("Create double pipe connection")
        self._fromPort = fromPort
        self._toPort = toPort
        self._editor = editor
        self._connection = None

    def redo(self):
        self._connection = _dpc.DoublePipeConnection(self._fromPort, self._toPort, self._editor)

    def undo(self):
        self._connection.deleteConn()
        self._connection = None



import dataclasses as _dc
import typing as _tp
import uuid as _uuid

import dataclasses_jsonschema as _dcj

import pytrnsys.utils.serialization as _ser

import trnsysGUI.connection.values as _values


@_dc.dataclass
class ConnectionModelVersion0(_ser.UpgradableJsonSchemaMixinVersion0):  # pylint: disable=too-many-instance-attributes
    ConnCID: int  # pylint: disable=invalid-name
    ConnDisplayName: str  # pylint: disable=invalid-name
    ConnID: int  # pylint: disable=invalid-name
    CornerPositions: _tp.List[_tp.Tuple[float, float]]  # pylint: disable=invalid-name
    FirstSegmentLabelPos: _tp.Tuple[float, float]  # pylint: disable=invalid-name
    FirstSegmentMassFlowLabelPos: _tp.Optional[_tp.Tuple[float, float]]  # pylint: disable=invalid-name
    GroupName: str  # pylint: disable=invalid-name
    PortFromID: int  # pylint: disable=invalid-name
    PortToID: int  # pylint: disable=invalid-name
    SegmentPositions: _tp.List[_tp.Tuple[float, float, float, float]]  # pylint: disable=invalid-name
    isVirtualConn: _tp.Optional[bool]
    trnsysID: int

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("7a15d665-f634-4037-b5af-3662b487a214")


@_dc.dataclass
class ConnectionModel(_ser.UpgradableJsonSchemaMixin):  # pylint: disable=too-many-instance-attributes
    connectionId: int
    name: str
    id: int  # pylint: disable=invalid-name
    segmentsCorners: _tp.List[_tp.Tuple[float, float]]
    labelPos: _tp.Tuple[float, float]
    massFlowLabelPos: _tp.Tuple[float, float]
    fromPortId: int
    toPortId: int
    trnsysId: int
    diameterInCm: _values.Value
    uValueInWPerM2K: _values.Value
    lengthInM: _values.Value

    @classmethod
    def from_dict(
        cls,
        data: _dcj.JsonDict,
        validate=True,
        validate_enums: bool = True,
    ) -> "ConnectionModel":
        data.pop(".__ConnectionDict__")
        connectionModel = super().from_dict(data, validate, validate_enums)
        return _tp.cast(ConnectionModel, connectionModel)

    def to_dict(
        self,
        omit_none: bool = True,
        validate: bool = False,
        validate_enums: bool = True,  # pylint: disable=duplicate-code
    ) -> _dcj.JsonDict:
        data = super().to_dict(omit_none, validate, validate_enums)
        data[".__ConnectionDict__"] = True
        return data

    @classmethod
    def getSupersededClass(cls) -> _tp.Type[_ser.UpgradableJsonSchemaMixinVersion0]:
        return ConnectionModelVersion0

    @classmethod
    def upgrade(cls, superseded: _ser.UpgradableJsonSchemaMixinVersion0) -> "ConnectionModel":
        assert isinstance(superseded, ConnectionModelVersion0)

        firstSegmentLabelPos = (
            superseded.SegmentPositions[0][0] + superseded.FirstSegmentLabelPos[0],
            superseded.SegmentPositions[0][1] + superseded.FirstSegmentLabelPos[1],
        )

        if superseded.FirstSegmentMassFlowLabelPos:
            firstSegmentMassFlowLabelPos = (
                superseded.SegmentPositions[0][0] + superseded.FirstSegmentMassFlowLabelPos[0],
                superseded.SegmentPositions[0][1] + superseded.FirstSegmentMassFlowLabelPos[1],
            )
        else:
            firstSegmentMassFlowLabelPos = firstSegmentLabelPos

        return ConnectionModel(
            superseded.ConnCID,
            superseded.ConnDisplayName,
            superseded.ConnID,
            superseded.CornerPositions,
            firstSegmentLabelPos,
            firstSegmentMassFlowLabelPos,
            superseded.PortFromID,
            superseded.PortToID,
            superseded.trnsysID,
            _values.DEFAULT_DIAMETER_IN_CM,
            _values.DEFAULT_U_VALUE_IN_W_PER_M2_K,
            _values.DEFAULT_LENGTH_IN_M,
        )

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("332cd663-684d-414a-b1ec-33fd036f0f17")



import dataclasses as _dc
import typing as _tp
import uuid as _uuid

import dataclasses_jsonschema as _dcj

import pytrnsys.utils.serialization as _ser
import trnsysGUI.blockItemModel as _bim
import trnsysGUI.createSinglePipePortItem as _cspi
import trnsysGUI.images as _img
import trnsysGUI.massFlowSolver.networkModel as _mfn
from trnsysGUI.BlockItem import BlockItem
from trnsysGUI.massFlowSolver import InternalPiping, MassFlowNetworkContributorMixin

_DEFAULT_MASS_FLOW_RATE = 500


class Pump(BlockItem, MassFlowNetworkContributorMixin):
    def __init__(self, trnsysType, parent, **kwargs):
        super().__init__(trnsysType, parent, **kwargs)
        self.w = 40
        self.h = 40

        self.massFlowRateInKgPerH = _DEFAULT_MASS_FLOW_RATE

        self.inputs.append(_cspi.createSinglePipePortItem("i", 0, self))
        self.outputs.append(_cspi.createSinglePipePortItem("o", 2, self))

        self.changeSize()

    def _getImageAccessor(self) -> _tp.Optional[_img.ImageAccessor]:
        return _img.PUMP_SVG

    def changeSize(self):
        width = self.w
        height = self.h

        delta = 20

        height = min(height, 20)
        width = min(width, 40)

        rect = self.label.boundingRect()
        labelWidth = rect.width()
        labelPosX = (width - labelWidth) / 2
        self.label.setPos(labelPosX, height)

        self.origInputsPos = [[0, delta]]
        self.origOutputsPos = [[width, delta]]
        self.inputs[0].setPos(self.origInputsPos[0][0], self.origInputsPos[0][1])
        self.outputs[0].setPos(self.origOutputsPos[0][0], self.origOutputsPos[0][1])

        self.updateFlipStateH(self.flippedH)
        self.updateFlipStateV(self.flippedV)

        self.inputs[0].side = (self.rotationN + 2 * self.flippedH) % 4
        self.outputs[0].side = (self.rotationN + 2 - 2 * self.flippedH) % 4

        return width, height

    def exportBlackBox(self):
        return "noBlackBoxOutput", []

    def exportPumpOutlets(self):
        equation = "T" + self.displayName + " = " + "T" + self.inputs[0].connectionList[0].displayName + "\n"
        equationNr = 1
        return equation, equationNr

    def exportMassFlows(self):
        equationNr = 1
        massFlowLine = f"Mfr{self.displayName} = {self.massFlowRateInKgPerH}\n"
        return massFlowLine, equationNr

    def getInternalPiping(self) -> InternalPiping:
        inputPort = _mfn.PortItem("input", _mfn.PortItemType.INPUT)
        outputPort = _mfn.PortItem("output", _mfn.PortItemType.OUTPUT)

        pump = _mfn.Pump(self.displayName, self.trnsysId, inputPort, outputPort)

        modelPortItemsToGraphicalPortItem = {inputPort: self.inputs[0], outputPort: self.outputs[0]}
        return InternalPiping([pump], modelPortItemsToGraphicalPortItem)

    def encode(self) -> _tp.Tuple[str, _dcj.JsonDict]:
        inputPortIds = [p.id for p in self.inputs]
        outputPortIds = [p.id for p in self.outputs]
        position = (self.pos().x(), self.pos().y())

        pumpModel = PumpModel(
            self.name,
            self.displayName,
            position,
            self.id,
            self.trnsysId,
            inputPortIds,
            outputPortIds,
            self.flippedH,
            self.flippedV,
            self.rotationN,
            self.massFlowRateInKgPerH
        )

        return "Block-", pumpModel.to_dict()

    def decode(self, i: _dcj.JsonDict, resBlockList) -> None:
        model = PumpModel.from_dict(i)

        self.setDisplayName(model.BlockDisplayName)
        self.setPos(float(model.blockPosition[0]), float(model.blockPosition[1]))
        self.id = model.Id
        self.trnsysId = model.trnsysId

        self.inputs[0].id = model.portsIdsIn[0]
        self.outputs[0].id = model.portsIdsOut[0]

        self.updateFlipStateH(model.flippedH)
        self.updateFlipStateV(model.flippedV)
        self.rotateBlockToN(model.rotationN)

        self.massFlowRateInKgPerH = model.massFlowRateInKgPerH

        resBlockList.append(self)


@_dc.dataclass
class PumpModel(_ser.UpgradableJsonSchemaMixin):  # pylint: disable=too-many-instance-attributes
    BlockName: str  # pylint: disable=invalid-name
    BlockDisplayName: str  # pylint: disable=invalid-name
    blockPosition: _tp.Tuple[float, float]
    Id: int  # pylint: disable=invalid-name,duplicate-code # 1
    trnsysId: int
    portsIdsIn: _tp.List[int]
    portsIdsOut: _tp.List[int]
    flippedH: bool
    flippedV: bool
    rotationN: int

    massFlowRateInKgPerH: float

    @classmethod
    def from_dict(
            cls,  # pylint: disable = duplicate-code 2
            data: _dcj.JsonDict,  # pylint: disable = duplicate-code 2
            validate=True,
            validate_enums: bool = True,
    ) -> "PumpModel":
        pumpModel = super().from_dict(data, validate, validate_enums)
        return _tp.cast(PumpModel, pumpModel)

    def to_dict(
        self,
        omit_none: bool = True,
        validate: bool = False,
        validate_enums: bool = True,  # pylint: disable=duplicate-code 2
    ) -> _dcj.JsonDict:
        data = super().to_dict(omit_none, validate, validate_enums)
        data[".__BlockDict__"] = True
        return data

    @classmethod
    def getSupersededClass(cls) -> _tp.Type[_ser.UpgradableJsonSchemaMixinVersion0]:
        return _bim.BlockItemModel

    @classmethod
    def upgrade(cls, superseded: _ser.UpgradableJsonSchemaMixinVersion0) -> "PumpModel":
        assert isinstance(superseded, _bim.BlockItemModel)

        return PumpModel(
            superseded.BlockName,
            superseded.BlockDisplayName,
            superseded.blockPosition,
            superseded.Id,
            superseded.trnsysId,
            superseded.portsIdsIn,
            superseded.portsIdsOut,
            superseded.flippedH,
            superseded.flippedV,
            superseded.rotationN,
            _DEFAULT_MASS_FLOW_RATE
        )

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("9552365f-ef11-4fac-ba35-644e94b54088")



import dataclasses as _dc
import typing as _tp

import dataclasses_jsonschema as _dcj


@_dc.dataclass
class Variable(_dcj.JsonSchemaMixin):
    name: str


Value = _tp.Union[Variable, float]

DEFAULT_DIAMETER_IN_CM = 2
DEFAULT_LENGTH_IN_M = 2.0
DEFAULT_U_VALUE_IN_W_PER_M2_K = 0.8333



from __future__ import annotations

import typing as _tp

import trnsysGUI.common as _com
import trnsysGUI.connection.singlePipeConnection as _spc
import trnsysGUI.massFlowSolver as _mfs
import trnsysGUI.massFlowSolver.networkModel as _mfn
import trnsysGUI.singlePipePortItem as _spi


def getReachableConnections(
        port: _spi.SinglePipePortItem,
        ignoreConnections: _tp.Set[_spc.SinglePipeConnection] = None,
) -> _tp.Set[_spc.SinglePipeConnection]:
    assert len(port.connectionList) <= 1

    portItems = {port}
    newPortItems, newConnections = _expandPortItemSetByOneLayer(portItems, ignoreConnections)
    while newPortItems != portItems:
        portItems = newPortItems
        newPortItems, newConnections = _expandPortItemSetByOneLayer(portItems, ignoreConnections)

    return newConnections


def _expandPortItemSetByOneLayer(
        portItems: set[_spi.SinglePipePortItem],
        ignoreConnections: _tp.Optional[set[_spc.SinglePipeConnection]] = None,
) -> _tp.Tuple[set[_spi.SinglePipePortItem], set[_spc.SinglePipeConnection]]:
    if ignoreConnections is None:
        ignoreConnections = set()

    connectedPortItems = [pi for pi in portItems if pi.connectionList]
    connections = [_com.getSingle(pi.connectionList) for pi in connectedPortItems]
    relevantConnections = {
        c
        for c in connections
        if isinstance(c, _spc.SinglePipeConnection) and c not in ignoreConnections
    }

    connectionPortItems = {p for c in relevantConnections for p in [c.fromPort, c.toPort]}

    internalPortItems = {mpi for p in portItems for mpi in getInternallyConnectedPortItems(p)}

    portItems = connectionPortItems | internalPortItems

    return portItems, relevantConnections


def getInternallyConnectedPortItems(
        port: _spi.SinglePipePortItem,
) -> _tp.Sequence[_spi.SinglePipePortItem]:
    contributor: _mfs.MassFlowNetworkContributorMixin = port.parent  # type: ignore[name-defined]
    internalPiping = contributor.getInternalPiping()

    graphicalPortItems = internalPiping.modelPortItemsToGraphicalPortItem
    startingNodes = internalPiping.openLoopsStartingNodes

    allInternallyConnectedModelPortItems = [_mfn.getConnectedRealNodesAndPortItems(sn).portItems for sn in
                                            startingNodes]

    allInternallyConnectedPortItems = []
    for internallyConnectedModelPortItems in allInternallyConnectedModelPortItems:
        internallyConnectedPortItems = []
        for modelPortItem in internallyConnectedModelPortItems:
            graphicalPortItem = graphicalPortItems[modelPortItem]
            if not isinstance(graphicalPortItem, _spi.SinglePipePortItem):
                continue

            internallyConnectedPortItems.append(graphicalPortItem)

        allInternallyConnectedPortItems.append(internallyConnectedPortItems)

    allIncidentInternallyConnectedPortItems = [pis for pis in allInternallyConnectedPortItems if port in pis]

    assert len(allIncidentInternallyConnectedPortItems) == 1

    incidentInternallyConnectedPortItems = allIncidentInternallyConnectedPortItems[0]

    return incidentInternallyConnectedPortItems



import dataclasses as _dc
import typing as _tp
import uuid as _uuid

import pytrnsys.utils.serialization as _ser


@_dc.dataclass
class Variable(_ser.UpgradableJsonSchemaMixinVersion0):
    name: str

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("3196024b-6775-42bc-92d2-e11d60c64bac")


@_dc.dataclass(frozen=True, eq=False)
class Fluid(_ser.UpgradableJsonSchemaMixinVersion0):
    name: str
    specificHeatCapacityInJPerKgK: _tp.Union[float, Variable]
    densityInKgPerM3: _tp.Union[float, Variable]

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("5b1d8eb8-e9c0-4c6a-b14f-0a7216b60132")


@_dc.dataclass
class HydraulicLoopVersion0(_ser.UpgradableJsonSchemaMixinVersion0):
    name: str
    hasUserDefinedName: bool
    fluidName: str
    connectionsTrnsysId: _tp.Sequence[int]

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("0047cab1-dd67-42ee-af45-5e5d30d97a92")


@_dc.dataclass
class HydraulicLoop(_ser.UpgradableJsonSchemaMixin):
    name: str
    hasUserDefinedName: bool
    fluidName: str
    useLoopWideDefaults: bool
    connectionsTrnsysId: _tp.Sequence[int]

    @classmethod
    def upgrade(cls, superseded: _ser.UpgradableJsonSchemaMixinVersion0) -> "HydraulicLoop":
        assert isinstance(superseded, HydraulicLoopVersion0)

        useLoopWideDefaults = False

        return HydraulicLoop(
            superseded.name,
            superseded.hasUserDefinedName,
            superseded.fluidName,
            useLoopWideDefaults,
            superseded.connectionsTrnsysId,
        )

    @classmethod
    def getSupersededClass(cls) -> _tp.Type[_ser.UpgradableJsonSchemaMixinVersion0]:
        return HydraulicLoopVersion0

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("990b8023-eb4b-408e-8d54-23caa5916b2a")



import PyQt5.QtWidgets as _qtw


def showErrorMessageBox(errorMessage: str, title: str = "Error") -> None:
    messageBox = _qtw.QMessageBox()
    messageBox.setWindowTitle(title)
    messageBox.setText(
        errorMessage
    )
    messageBox.setStandardButtons(_qtw.QMessageBox.Ok)
    messageBox.exec()



import dataclasses as _dc
import logging as _log
import pathlib as _pl
import shutil as _su
import subprocess as _sp
import sys as _sys
import typing as _tp

import PyQt5.QtWidgets as _qtw
import pandas as _pd
import pytest as _pt

import trnsysGUI.diagram.Editor as _de
import trnsysGUI.storageTank.widget as _stw

_DATA_DIR = _pl.Path(__file__).parent / "data"


@_dc.dataclass
class _Project:
    projectName: str
    testCasesDirName: str
    shallCopyFolderFromExamples: bool

    @staticmethod
    def createForTestProject(projectName: str) -> "_Project":
        return _Project(projectName, "tests", False)

    @staticmethod
    def createForExampleProject(projectName: str) -> "_Project":
        return _Project(projectName, "examples", True)

    @property
    def testId(self) -> str:
        return f"{self.projectName} [in {self.testCasesDirName}]"


def getProjects() -> _tp.Iterable[_Project]:
    yield _Project.createForExampleProject("TRIHP_dualSource")

    yield from getTestProjects()


def getTestProjects() -> _tp.Iterable[_Project]:
    testProjectTestCasesDir = _DATA_DIR / "tests"
    testProjectTestCaseDirPaths = [tc for tc in testProjectTestCasesDir.iterdir() if tc.name != "README.txt"]
    for testProjectTestCaseDirPath in testProjectTestCaseDirPaths:
        projectName = testProjectTestCaseDirPath.name
        yield _Project.createForTestProject(projectName)


TEST_CASES = [_pt.param(p, id=p.testId) for p in getProjects()]


class TestEditor:
    @_pt.mark.parametrize("project", TEST_CASES)
    def testStorageAndHydraulicExports(self, project: _Project, request: _pt.FixtureRequest) -> None:
        helper = _Helper(project)
        helper.setup()

        # The following line is required otherwise QT will crash
        application = _qtw.QApplication([])

        def quitApplication():
            application.quit()

        request.addfinalizer(quitApplication)

        projectFolderPath = helper.actualProjectFolderPath

        self._exportHydraulic(projectFolderPath, _format="mfs")
        mfsDdckRelativePath = f"{project.projectName}_mfs.dck"
        helper.ensureFilesAreEqual(mfsDdckRelativePath)

        self._exportHydraulic(projectFolderPath, _format="ddck")
        hydraulicDdckRelativePath = "ddck/hydraulic/hydraulic.ddck"
        helper.ensureFilesAreEqual(hydraulicDdckRelativePath)

        storageTankNames = self._exportStorageTanksAndGetNames(projectFolderPath)
        for storageTankName in storageTankNames:
            ddckFileRelativePath = f"ddck/{storageTankName}/{storageTankName}.ddck"
            helper.ensureFilesAreEqual(ddckFileRelativePath)

            ddcxFileRelativePath = f"ddck/{storageTankName}/{storageTankName}.ddcx"
            helper.ensureFilesAreEqual(ddcxFileRelativePath)

    def _exportStorageTanksAndGetNames(self, projectFolderPath: _pl.Path) -> _tp.Sequence[str]:
        editor = self._createEditor(projectFolderPath)
        storageTanks = [o for o in editor.trnsysObj if isinstance(o, _stw.StorageTank)]  # pylint: disable=no-member

        for storageTank in storageTanks:
            storageTank.exportDck()

        storageTankNames = [t.displayName for t in storageTanks]

        return storageTankNames

    @_pt.mark.linux_ci
    @_pt.mark.parametrize("project", TEST_CASES)
    def testMassFlowSolver(self, project: _Project, request: _pt.FixtureRequest) -> None:
        helper = _Helper(project)
        helper.setup()

        # The following line is required otherwise QT will crash
        application = _qtw.QApplication([])

        def quitApplication():
            application.quit()

        request.addfinalizer(quitApplication)

        projectFolderPath = helper.actualProjectFolderPath

        self._exportMassFlowSolverDeckAndRunTrnsys(projectFolderPath)

        massFlowRatesPrintFile = f"{project.projectName}_Mfr.prt"
        helper.ensureCSVsAreEqual(massFlowRatesPrintFile)

        temperaturesPintFileName = f"{project.projectName}_T.prt"
        helper.ensureCSVsAreEqual(temperaturesPintFileName)

    def _exportMassFlowSolverDeckAndRunTrnsys(self, projectFolderPath):
        exportedFilePath = self._exportHydraulic(projectFolderPath, _format="mfs")

        trnsysExePath = _pl.PureWindowsPath(r"C:\TRNSYS18\Exe\TrnExe.exe")

        runningOnWindows = _sys.platform.startswith("win32")
        if runningOnWindows:
            _sp.run(
                [str(trnsysExePath), str(exportedFilePath), "/H"], check=True
            )
            return

        # Running on Linux using Wine
        zDrivePath = _pl.PureWindowsPath("Z:")
        exportedFileWindowswPath = zDrivePath / exportedFilePath

        _sp.run(
            ["wine", str(trnsysExePath), str(exportedFileWindowswPath), "/H"], check=True
        )

    @classmethod
    def _exportHydraulic(cls, projectFolderPath, *, _format) -> str:
        editor = cls._createEditor(projectFolderPath)
        exportedFilePath = editor.exportHydraulics(exportTo=_format)
        return exportedFilePath

    @staticmethod
    def _createEditor(projectFolderPath):
        logger = _log.Logger("root")
        editor = _de.Editor(
            parent=None,
            projectFolder=str(projectFolderPath),
            jsonPath=None,
            loadValue="load",
            logger=logger,
        )
        return editor


class _Helper:
    def __init__(
        self,
        project: _Project,
    ):
        self._project = project

        testCasesFolderPath = _DATA_DIR / self._project.testCasesDirName

        # project name is used twice because the project file must always be contained in a folder
        # of the same name
        self.actualProjectFolderPath = testCasesFolderPath / self._project.projectName / self._project.projectName

        self._expectedProjectFolderPath = testCasesFolderPath / self._project.projectName / "expected"

    def setup(self):
        if self._project.shallCopyFolderFromExamples:
            self._copyExampleToTestInputFolder()

        self._removeGeneratedFiles()

    def _removeGeneratedFiles(self):
        for child in self.actualProjectFolderPath.iterdir():
            if child.name in ("ddck", f"{self._project.projectName}.json"):
                continue

            if child.is_dir():
                _su.rmtree(child)
            else:
                child.unlink()

    def ensureFilesAreEqual(self, fileRelativePathAsString: str) -> None:
        actualFilePath, expectedFilePath = self._getActualAndExpectedFilePath(fileRelativePathAsString)
        actualContent = actualFilePath.read_text()
        expectedContent = expectedFilePath.read_text()

        assert actualContent == expectedContent

    def _getActualAndExpectedFilePath(self, fileRelativePathAsString):
        fileRelativePath = _pl.Path(fileRelativePathAsString)
        actualFilePath = self.actualProjectFolderPath / fileRelativePath
        expectedFilePath = self._expectedProjectFolderPath / fileRelativePath
        return actualFilePath, expectedFilePath

    def ensureCSVsAreEqual(self, fileRelativePathAsString: str, absoluteTolerance: float = 1e-10) -> None:
        actualFilePath, expectedFilePath = self._getActualAndExpectedFilePath(fileRelativePathAsString)

        actualDf: _pd.DataFrame = _pd.read_csv(actualFilePath, delim_whitespace=True)
        expectedDf: _pd.DataFrame = _pd.read_csv(expectedFilePath, delim_whitespace=True)

        assert actualDf.shape == expectedDf.shape

        maxAbsoluteDifference = (actualDf - expectedDf).max().max()

        assert maxAbsoluteDifference <= absoluteTolerance

    def _copyExampleToTestInputFolder(self):
        if self.actualProjectFolderPath.exists():
            _su.rmtree(self.actualProjectFolderPath)

        pytrnsysGuiDir = _pl.Path(__file__).parents[3]
        exampleFolderPath = pytrnsysGuiDir / "data" / "examples" / self._project.projectName

        _su.copytree(exampleFolderPath, self.actualProjectFolderPath)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'trnsysGUI/hydraulicLoops/_dialogs/merge/_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.15.3
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_mergeLoopsDialog(object):
    def setupUi(self, mergeLoopsDialog):
        mergeLoopsDialog.setObjectName("mergeLoopsDialog")
        mergeLoopsDialog.resize(340, 128)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(mergeLoopsDialog.sizePolicy().hasHeightForWidth())
        mergeLoopsDialog.setSizePolicy(sizePolicy)
        mergeLoopsDialog.setMinimumSize(QtCore.QSize(0, 0))
        mergeLoopsDialog.setMaximumSize(QtCore.QSize(16777215, 128))
        self.verticalLayout = QtWidgets.QVBoxLayout(mergeLoopsDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.differingFluidsWarningWidget = QtWidgets.QWidget(mergeLoopsDialog)
        self.differingFluidsWarningWidget.setMinimumSize(QtCore.QSize(30, 30))
        self.differingFluidsWarningWidget.setStyleSheet("image: url(:/tango-icons/status/tango-icon-theme-0.8.90/scalable/status/dialog-warning.svg);")
        self.differingFluidsWarningWidget.setObjectName("differingFluidsWarningWidget")
        self.horizontalLayout_2.addWidget(self.differingFluidsWarningWidget)
        self.differingFluidsWarningLabel = QtWidgets.QLabel(mergeLoopsDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.differingFluidsWarningLabel.sizePolicy().hasHeightForWidth())
        self.differingFluidsWarningLabel.setSizePolicy(sizePolicy)
        self.differingFluidsWarningLabel.setWordWrap(True)
        self.differingFluidsWarningLabel.setObjectName("differingFluidsWarningLabel")
        self.horizontalLayout_2.addWidget(self.differingFluidsWarningLabel)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.loopNameLabel = QtWidgets.QLabel(mergeLoopsDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.loopNameLabel.sizePolicy().hasHeightForWidth())
        self.loopNameLabel.setSizePolicy(sizePolicy)
        self.loopNameLabel.setObjectName("loopNameLabel")
        self.horizontalLayout.addWidget(self.loopNameLabel)
        self.nameComboBox = QtWidgets.QComboBox(mergeLoopsDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.nameComboBox.sizePolicy().hasHeightForWidth())
        self.nameComboBox.setSizePolicy(sizePolicy)
        self.nameComboBox.setMinimumSize(QtCore.QSize(100, 0))
        self.nameComboBox.setEditable(True)
        self.nameComboBox.setObjectName("nameComboBox")
        self.horizontalLayout.addWidget(self.nameComboBox)
        self.invalidNameWarningWidget = QtWidgets.QWidget(mergeLoopsDialog)
        self.invalidNameWarningWidget.setMinimumSize(QtCore.QSize(30, 30))
        self.invalidNameWarningWidget.setStyleSheet("image: url(:/tango-icons/status/tango-icon-theme-0.8.90/scalable/status/dialog-warning.svg);")
        self.invalidNameWarningWidget.setObjectName("invalidNameWarningWidget")
        self.horizontalLayout.addWidget(self.invalidNameWarningWidget)
        spacerItem = QtWidgets.QSpacerItem(30, 20, QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.fluidLabel = QtWidgets.QLabel(mergeLoopsDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fluidLabel.sizePolicy().hasHeightForWidth())
        self.fluidLabel.setSizePolicy(sizePolicy)
        self.fluidLabel.setObjectName("fluidLabel")
        self.horizontalLayout.addWidget(self.fluidLabel)
        self.fluidComboBox = QtWidgets.QComboBox(mergeLoopsDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fluidComboBox.sizePolicy().hasHeightForWidth())
        self.fluidComboBox.setSizePolicy(sizePolicy)
        self.fluidComboBox.setMinimumSize(QtCore.QSize(100, 0))
        self.fluidComboBox.setObjectName("fluidComboBox")
        self.horizontalLayout.addWidget(self.fluidComboBox)
        self.noFluidSelectedWarningWidget = QtWidgets.QWidget(mergeLoopsDialog)
        self.noFluidSelectedWarningWidget.setMinimumSize(QtCore.QSize(30, 30))
        self.noFluidSelectedWarningWidget.setStyleSheet("image: url(:/tango-icons/status/tango-icon-theme-0.8.90/scalable/status/dialog-warning.svg);")
        self.noFluidSelectedWarningWidget.setObjectName("noFluidSelectedWarningWidget")
        self.horizontalLayout.addWidget(self.noFluidSelectedWarningWidget)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem2)
        self.applyButton = QtWidgets.QPushButton(mergeLoopsDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.applyButton.sizePolicy().hasHeightForWidth())
        self.applyButton.setSizePolicy(sizePolicy)
        self.applyButton.setObjectName("applyButton")
        self.horizontalLayout_5.addWidget(self.applyButton)
        self.abortButton = QtWidgets.QPushButton(mergeLoopsDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.abortButton.sizePolicy().hasHeightForWidth())
        self.abortButton.setSizePolicy(sizePolicy)
        self.abortButton.setObjectName("abortButton")
        self.horizontalLayout_5.addWidget(self.abortButton)
        self.verticalLayout.addLayout(self.horizontalLayout_5)

        self.retranslateUi(mergeLoopsDialog)
        QtCore.QMetaObject.connectSlotsByName(mergeLoopsDialog)

    def retranslateUi(self, mergeLoopsDialog):
        _translate = QtCore.QCoreApplication.translate
        mergeLoopsDialog.setWindowTitle(_translate("mergeLoopsDialog", "Hydraulic loops merge"))
        self.differingFluidsWarningLabel.setText(_translate("mergeLoopsDialog", "The fluids of the loops you are about to merge are different. Please choose the fluid for the merged loop or abort."))
        self.loopNameLabel.setText(_translate("mergeLoopsDialog", "Name"))
        self.fluidLabel.setText(_translate("mergeLoopsDialog", "Fluid"))
        self.noFluidSelectedWarningWidget.setToolTip(_translate("mergeLoopsDialog", "<html><head/><body><p>A fluid must be selected</p></body></html>"))
        self.applyButton.setText(_translate("mergeLoopsDialog", "Apply"))
        self.abortButton.setText(_translate("mergeLoopsDialog", "Abort"))



__all__ = [
    "NodeType",
    "NodeBase",
    "PortItem",
    "RealNodeBase",
    "OneNeighbourBase",
    "Source",
    "Sink",
    "TwoNeighboursBase",
    "Pipe",
    "Pump",
    "ThreeNeighboursBase",
    "TeePiece",
    "Diverter",
    "getConnectedNodes",
    "getConnectedRealNodesAndPortItems",
]

import abc as _abc
import dataclasses as _dc
import enum as _enum
import typing as _tp
import collections.abc as _cabc


class NodeType(_enum.IntEnum):
    PIPE = 0
    PUMP = 1
    TEE_PIECE = 2
    DIVERTER = 3
    SOURCE = 4
    SINK = 5


@_dc.dataclass(eq=False)
class InputVariable:
    UNDEFINED_INPUT_VARIABLE_VALUE = "0,0"
    name: str


@_dc.dataclass(eq=False)
class OutputVariable:
    name: str


OutputVariableIndex = _tp.Literal[1, 2, 3]

MAX_N_OUTPUT_VARIABLES_PER_NODE = 3

OutputVariables = _tp.Tuple[_tp.Optional[OutputVariable], _tp.Optional[OutputVariable], _tp.Optional[OutputVariable]]


class NodeBase(_abc.ABC):
    @_abc.abstractmethod
    def getNeighbours(self) -> _tp.Sequence["NodeBase"]:
        raise NotImplementedError()


class PortItemType(_enum.Enum):
    INPUT = "Input"
    OUTPUT = "Output"


@_dc.dataclass(eq=False, repr=False)
class PortItem(NodeBase, _abc.ABC):
    name: str
    type: PortItemType

    def getNeighbours(self) -> _tp.Sequence[NodeBase]:
        return []

    def canOverlapWith(self, other: "PortItem") -> bool:
        """Can this model port item overlap with `other` under a graphical port item?"""
        return other != self


@_dc.dataclass(eq=False)  # type: ignore[misc]
class RealNodeBase(NodeBase, _abc.ABC):
    name: str
    trnsysId: int

    @_abc.abstractmethod
    def getNodeType(self) -> NodeType:
        raise NotImplementedError()

    def getInputVariable(self) -> _tp.Optional[InputVariable]:  # pylint: disable=no-self-use
        return None

    def getOutputVariables(self) -> OutputVariables:
        return (
            self._getOutputVariable(1, "A"),
            self._getOutputVariable(2, "B"),
            self._getOutputVariable(3, "C"),
        )

    def _getOutputVariable(self, variableIndex: OutputVariableIndex, postfix: str) -> _tp.Optional[OutputVariable]:
        if variableIndex > self._getNumberOfOutputVariables():
            return None

        return OutputVariable(f"Mfr{self.name}_{postfix}")

    @_abc.abstractmethod
    def _getNumberOfOutputVariables(self) -> OutputVariableIndex:
        raise NotImplementedError()


@_dc.dataclass(eq=False)  # type: ignore[misc]
class OneNeighbourBase(RealNodeBase, _abc.ABC):
    neighbour: NodeBase

    def getNeighbours(self) -> _tp.Sequence[NodeBase]:
        return [self.neighbour]


class Source(OneNeighbourBase):
    def getNodeType(self) -> NodeType:
        return NodeType.SOURCE

    def _getNumberOfOutputVariables(self) -> OutputVariableIndex:
        return 2

    def getInputVariable(self) -> _tp.Optional[InputVariable]:
        return InputVariable(f"Mfr{self.name}")


class Sink(OneNeighbourBase):
    def getNodeType(self) -> NodeType:
        return NodeType.SINK

    def _getNumberOfOutputVariables(self) -> OutputVariableIndex:
        return 2


@_dc.dataclass(eq=False)  # type: ignore[misc]
class TwoNeighboursBase(RealNodeBase, _abc.ABC):
    fromNode: NodeBase
    toNode: NodeBase

    def getNeighbours(self) -> _tp.Sequence[NodeBase]:
        return [self.fromNode, self.toNode]

    def _getNumberOfOutputVariables(self) -> OutputVariableIndex:
        return 2


class Pipe(TwoNeighboursBase):
    def getNodeType(self) -> NodeType:
        return NodeType.PIPE

    @property
    def fromPort(self) -> PortItem:
        assert isinstance(self.fromNode, PortItem)
        return self.fromNode

    @property
    def toPort(self) -> PortItem:
        assert isinstance(self.toNode, PortItem)
        return self.toNode


class Pump(TwoNeighboursBase):
    def getNodeType(self) -> NodeType:
        return NodeType.PUMP

    def getInputVariable(self) -> _tp.Optional[InputVariable]:
        return InputVariable(f"Mfr{self.name}")


@_dc.dataclass(eq=False)  # type: ignore[misc]
class ThreeNeighboursBase(RealNodeBase, _abc.ABC):
    node1: NodeBase
    node2: NodeBase
    node3: NodeBase

    def getNeighbours(self) -> _tp.Sequence[NodeBase]:
        return [self.node1, self.node2, self.node3]

    def _getNumberOfOutputVariables(self) -> OutputVariableIndex:
        return 3


class Diverter(ThreeNeighboursBase):
    def getNodeType(self) -> NodeType:
        return NodeType.DIVERTER

    def getInputVariable(self) -> _tp.Optional[InputVariable]:
        return InputVariable(f"xFrac{self.name}")


class TeePiece(ThreeNeighboursBase):
    def getNodeType(self) -> NodeType:
        return NodeType.TEE_PIECE


def getConnectedNodes(startingNode: NodeBase) -> _tp.Sequence[NodeBase]:
    nodes = {startingNode}
    neighbours = _getNeighboursForNodes(nodes)
    while neighbours:
        nodes |= neighbours
        neighbours = _getNeighboursForNodes(nodes)

    return list(nodes)


def _getNeighboursForNodes(nodes: _cabc.Set[NodeBase]) -> _cabc.Set[NodeBase]:
    neighbours = {neighbour for node in nodes for neighbour in node.getNeighbours() if neighbour not in nodes}
    return neighbours


@_dc.dataclass
class RealNodesAndPortItems:
    realNodes: _tp.Sequence[RealNodeBase]
    portItems: _tp.Sequence[PortItem]


def getConnectedRealNodesAndPortItems(startingNode: NodeBase) -> RealNodesAndPortItems:
    connectedNodes = getConnectedNodes(startingNode)
    realNodes = [n for n in connectedNodes if isinstance(n, RealNodeBase)]
    portItems = [n for n in connectedNodes if isinstance(n, PortItem)]
    return RealNodesAndPortItems(realNodes, portItems)



# pylint: disable = invalid-name

from __future__ import annotations

import typing as _tp

import PyQt5.QtCore as _qtc
from PyQt5.QtWidgets import (
    QLabel,
    QLineEdit,
    QGridLayout,
    QHBoxLayout,
    QListWidget,
    QPushButton,
    QSpacerItem,
    QVBoxLayout,
    QRadioButton,
    QDialog,
    QTabWidget,
    QWidget,
    QMessageBox,
    QListWidgetItem,
)

import trnsysGUI.directPortPair as _dpp
import trnsysGUI.modifyRelativeHeightsDialog as _mhd
import trnsysGUI.storageTank.side as _sd

if _tp.TYPE_CHECKING:
    import trnsysGUI.storageTank.widget as _st
    import trnsysGUI.diagram.Editor as _ed


class ConfigureStorageDialog(QDialog):  # pylint: disable = too-many-instance-attributes
    WIDTH_INCREMENT = 10
    HEIGHT_INCREMENT = 100

    def __init__(self, storage: _st.StorageTank, editor: _ed.Editor):  # type: ignore[name-defined]
        # pylint: disable = too-many-locals, too-many-statements
        super().__init__(editor)
        self._editor = editor
        self.storage = storage
        self.n = 0  # pylint: disable = invalid-name
        self.m = 0  # pylint: disable = invalid-name

        spacerHeight = 15

        self.tabs = QTabWidget()
        self.tab1 = QWidget()
        self.tab2 = QWidget()

        self.tabs.addTab(self.tab1, "Heat exchangers")
        self.tabs.addTab(self.tab2, "Direct ports")

        h0 = QHBoxLayout()  # pylint: disable = invalid-name
        description = QLabel("Please configure the storage tank:")
        exportButton = QPushButton("Export ddck")
        exportButton.clicked.connect(self.storage.exportDck)

        h0.addWidget(description)
        h0.addWidget(exportButton)

        tankNameLabel = QLabel()
        tankNameLabel.setText("<b>Tank name: </b>")
        self.lineEdit = QLineEdit(self.storage.label.toPlainText())

        gridLayout = QGridLayout()
        hxsLabel = QLabel("<b>Heat Exchangers </b>")
        gridLayout.addWidget(hxsLabel, 0, 0, 1, 2)

        qhbL = QHBoxLayout()

        self.leftHeatExchangersItemListWidget = QListWidget()
        qhbL.addWidget(self.leftHeatExchangersItemListWidget)

        self.rightHeatExchangersItemListWidget = QListWidget()
        qhbL.addWidget(self.rightHeatExchangersItemListWidget)

        offsetLabel = QLabel("Height offsets in percent")
        offsetLeILabel = QLabel("Input (upper port): ")
        offsetLeOLabel = QLabel("Output (lower port):")
        self.offsetLeI = QLineEdit("0")
        self.offsetLeO = QLineEdit("0")
        self.lButton = QRadioButton("Left side")
        self.rButton = QRadioButton("Right side")

        gridLayout.addWidget(offsetLabel, 3, 0, 1, 1)
        gridLayout.addWidget(offsetLeILabel, 4, 0, 1, 1)
        gridLayout.addWidget(self.offsetLeI, 5, 0, 1, 3)
        gridLayout.addWidget(offsetLeOLabel, 6, 0, 1, 1)
        gridLayout.addWidget(self.offsetLeO, 7, 0, 1, 3)

        self.hxNameLe = QLineEdit()
        self.hxNameLe.setPlaceholderText("Enter a name...")
        gridLayout.addWidget(self.hxNameLe, 8, 0, 1, 3)

        gridLayout.addWidget(self.lButton, 9, 0, 1, 1)
        gridLayout.addWidget(self.rButton, 9, 2, 1, 1)

        addButton = QPushButton("Add...")
        addButton.clicked.connect(self.addHx)
        gridLayout.addWidget(addButton, 10, 0, 1, 1)
        removeButton = QPushButton("Remove...")
        removeButton.clicked.connect(self.removeHxL)
        removeButton.clicked.connect(self.removeHxR)
        gridLayout.addWidget(removeButton, 10, 1, 1, 1)
        modifyButton = QPushButton("Modify")
        modifyButton.clicked.connect(self.modifyHx)
        gridLayout.addWidget(modifyButton, 10, 2, 1, 1)
        spaceHx = QSpacerItem(self.width(), spacerHeight)
        gridLayout.addItem(spaceHx, 11, 0, 1, 2)

        manPortLay = QVBoxLayout()
        qhbL2 = QHBoxLayout()

        self._leftDirectPortPairsItemListWidget = QListWidget()
        qhbL2.addWidget(self._leftDirectPortPairsItemListWidget)

        self._rightDirectPortPairsItemListWidget = QListWidget()
        qhbL2.addWidget(self._rightDirectPortPairsItemListWidget)

        manPortLay.addLayout(qhbL2)

        manPortLabel = QLabel("<b>Set port manually</b>")
        manPortLabel2 = QLabel("Enter height in percent: ")
        portlabelUpper = QLabel("Inlet")
        self.manPortLeI = QLineEdit("0")
        portlabelLower = QLabel("Outlet")
        self.manPortLeO = QLineEdit("0")

        qhbl3 = QHBoxLayout()
        self.manlButton = QRadioButton("Left side")
        self.manrButton = QRadioButton("Right side")
        qhbl3.addWidget(self.manlButton)
        qhbl3.addWidget(self.manrButton)

        self.manAddButton = QPushButton("Add (manual) ports")
        self.manAddButton.clicked.connect(self.addPortPair)

        self.manRemovebutton = QPushButton("Remove ports")
        self.manRemovebutton.clicked.connect(self.removePortPairLeft)
        self.manRemovebutton.clicked.connect(self.removePortPairRight)

        self.modifyPortButton = QPushButton("Modify")
        self.modifyPortButton.clicked.connect(self.modifyPort)

        addRemoveButtons = QHBoxLayout()
        addRemoveButtons.addWidget(self.manAddButton)
        addRemoveButtons.addWidget(self.manRemovebutton)
        addRemoveButtons.addWidget(self.modifyPortButton)

        manPortLay.addWidget(manPortLabel)
        manPortLay.addWidget(manPortLabel2)
        manPortLay.addWidget(portlabelUpper)
        manPortLay.addWidget(self.manPortLeI)
        manPortLay.addWidget(portlabelLower)
        manPortLay.addWidget(self.manPortLeO)
        manPortLay.addLayout(qhbl3)
        manPortLay.addLayout(addRemoveButtons)

        increaseSizeButton = QPushButton("Increase size")
        decreaseSizeButton = QPushButton("Decrease size")
        self.okButton = QPushButton("OK")
        self.cancelButton = QPushButton("Cancel")

        increaseSizeButton.clicked.connect(self.incrSize)
        decreaseSizeButton.clicked.connect(self.decrSize)
        self.okButton.clicked.connect(self.acceptedEdit)
        self.cancelButton.clicked.connect(self.cancel)

        layout = QVBoxLayout()
        layout.addLayout(h0)
        layout.addWidget(tankNameLabel)
        layout.addWidget(self.lineEdit)

        t1Layout = QVBoxLayout()
        t1Layout.addLayout(qhbL)
        t1Layout.addLayout(gridLayout)

        self.tab1.setLayout(t1Layout)

        self.tab2.setLayout(manPortLay)

        layout.addWidget(self.tabs)

        layout.addWidget(increaseSizeButton)
        layout.addWidget(decreaseSizeButton)
        layout.addWidget(self.okButton)
        layout.addWidget(self.cancelButton)

        self.setLayout(layout)

        self._loadHeatExchangers()
        self._loadDirectPortPairs()

        # This is to ensure that only one list element is selected
        self.rightHeatExchangersItemListWidget.setSelectionMode(QListWidget.SelectionMode(1))
        self.leftHeatExchangersItemListWidget.setSelectionMode(QListWidget.SelectionMode(1))
        self.rightHeatExchangersItemListWidget.clicked.connect(self.listWRClicked)
        self.leftHeatExchangersItemListWidget.clicked.connect(self.listWLClicked)

        self._rightDirectPortPairsItemListWidget.setSelectionMode(QListWidget.SelectionMode(1))
        self._leftDirectPortPairsItemListWidget.setSelectionMode(QListWidget.SelectionMode(1))
        self._rightDirectPortPairsItemListWidget.clicked.connect(self.listWR2Clicked)
        self._leftDirectPortPairsItemListWidget.clicked.connect(self.listWL2Clicked)

        self.show()

    def _loadHeatExchangers(self):
        self.leftHeatExchangersItemListWidget.clear()
        self.rightHeatExchangersItemListWidget.clear()

        for heatExchanger in self.storage.heatExchangers:
            itemText = self._getHeatExchangerListItemText(heatExchanger)
            item = QListWidgetItem(itemText)
            item.setData(_qtc.Qt.UserRole, heatExchanger)

            if heatExchanger.sSide == 0:
                self.leftHeatExchangersItemListWidget.addItem(item)
            else:
                self.rightHeatExchangersItemListWidget.addItem(item)

    @staticmethod
    def _getHeatExchangerListItemText(h):  # pylint: disable = invalid-name
        return (
            f"{h.displayName}, y_offset = {int(h.relativeInputHeight * 100)}% to {int(h.relativeOutputHeight * 100)}%"
        )

    def _loadDirectPortPairs(self):
        self._leftDirectPortPairsItemListWidget.clear()
        self._rightDirectPortPairsItemListWidget.clear()

        directPortPair: _dpp.DirectPortPair
        for directPortPair in self.storage.directPortPairs:
            itemText = self._getDirectPortPairListItemText(directPortPair)
            item = QListWidgetItem(itemText)
            item.setData(_qtc.Qt.UserRole, directPortPair)

            if directPortPair.side.isLeft:
                self._leftDirectPortPairsItemListWidget.addItem(item)
            else:
                self._rightDirectPortPairsItemListWidget.addItem(item)

    @staticmethod
    def _getDirectPortPairListItemText(directPortPair: _dpp.DirectPortPair):
        return (
            f"Port pair from {int(directPortPair.relativeInputHeight * 100)}% to "
            f"{int(directPortPair.relativeOutputHeight * 100)}%"
        )

    def listWLClicked(self):
        self.rightHeatExchangersItemListWidget.clearSelection()

    def listWRClicked(self):
        self.leftHeatExchangersItemListWidget.clearSelection()

    def listWL2Clicked(self):
        self._rightDirectPortPairsItemListWidget.clearSelection()

    def listWR2Clicked(self):
        self._leftDirectPortPairsItemListWidget.clearSelection()

    def addHx(self):
        """
        Checks whether the inputs are in the correct range (0,100) and in order, and calls the function for creating a
        HeatExchanger on the respective side.

        Returns
        -------
        """
        try:
            _inputPercentageHeight = float(self.offsetLeI.text())
        except ValueError:
            self._editor.logger.warning('HX input height is not a number.')
            return

        try:
            _outputPercentageHeight = float(self.offsetLeO.text())
        except ValueError:
            self._editor.logger.warning('HX output height is not a number.')
            return

        if (
                self.minOffsetDistance()
                and _inputPercentageHeight > _outputPercentageHeight
                and self.offsetsInRange()
        ):
            if self.rButton.isChecked():
                self._editor.logger.debug('Adding HX on righthand side.')
                self._addHxR()
            elif self.lButton.isChecked():
                self._editor.logger.debug('Adding HX on lefthand side.')
                self._addHxL()
            else:
                self._editor.logger.warning('No side selected for heat exchanger.')
                return
        else:
            msgb = QMessageBox()
            msgb.setText(
                "At least 20% of difference and larger top port than bottom port needed and valid range (0, 100)"
            )
            msgb.exec_()

    def minOffsetDistance(self):
        return abs(float(self.offsetLeI.text()) - float(self.offsetLeO.text())) >= 5

    def offsetsInRange(self):
        return (0 <= float(self.offsetLeI.text()) <= 100) and (0 <= float(self.offsetLeO.text()) <= 100)

    def _addHxL(self):
        self._addHeatExchanger(_sd.Side.LEFT)

    def _addHxR(self):
        self._addHeatExchanger(_sd.Side.RIGHT)

    def _addHeatExchanger(self, side: _sd.Side):
        name = self.hxNameLe.text()
        if not name:
            messageBox = QMessageBox()
            messageBox.setText("Please specify the name of the heat exchanger that you want to add.")
            messageBox.exec_()
            return

        relativeInputHeight = float(self.offsetLeI.text()) / 100
        relativeOutputHeight = float(self.offsetLeO.text()) / 100

        trnsysId = self._editor.idGen.getTrnsysID()  # type: ignore[attr-defined]
        self.storage.addHeatExchanger(name, trnsysId, side, relativeInputHeight, relativeOutputHeight)

        self._loadHeatExchangers()

    def addPortPair(self):
        try:
            _inputPortPercentageHeight = float(self.manPortLeI.text())
        except ValueError:
            self._editor.logger.warning('Input port height is not a number.')
            return

        try:
            _outputPortPercentageHeight = float(self.manPortLeO.text())
        except ValueError:
            self._editor.logger.warning('Output port height is not a number.')
            return

        if max(_inputPortPercentageHeight, _outputPortPercentageHeight) >= 100 or min(_inputPortPercentageHeight,
                                                                                      _outputPortPercentageHeight) <= 0:
            messageBox = QMessageBox()
            messageBox.setText(
                'Ports need to be on the tank, please make sure the port heights are within (0 %, 100 %).')
            messageBox.exec_()
            return

        trnsysId = self._editor.idGen.getTrnsysID()

        if self.manlButton.isChecked():
            _pairSide = _sd.Side.LEFT
        elif self.manrButton.isChecked():
            _pairSide = _sd.Side.RIGHT
        else:
            self._editor.logger.warning('No side selected for port pair.')
            return

        self.storage.addDirectPortPair(
            trnsysId,
            _pairSide,
            _inputPortPercentageHeight / 100,
            _outputPortPercentageHeight / 100,
            self.storage.h,
        )

        self._loadDirectPortPairs()

    def removePortPairLeft(self):
        self._removeSelectedPortPairs(self._leftDirectPortPairsItemListWidget)

    def removePortPairRight(self):
        self._removeSelectedPortPairs(self._rightDirectPortPairsItemListWidget)

    def _removeSelectedPortPairs(self, directPortPairsListWidget):
        for selectedItem in directPortPairsListWidget.selectedItems():
            selectedDirectPortPair = selectedItem.data(_qtc.Qt.UserRole)

            self.storage.directPortPairs.remove(selectedDirectPortPair)

            row = directPortPairsListWidget.row(selectedItem)
            directPortPairsListWidget.takeItem(row)

            while len(selectedDirectPortPair.fromPort.connectionList) > 0:
                selectedDirectPortPair.fromPort.connectionList[0].deleteConn()

            while len(selectedDirectPortPair.toPort.connectionList) > 0:
                selectedDirectPortPair.toPort.connectionList[0].deleteConn()

            self.storage.inputs.remove(selectedDirectPortPair.fromPort)
            self.storage.outputs.remove(selectedDirectPortPair.toPort)

            self.storage.parent.scene().removeItem(selectedDirectPortPair.fromPort)
            self.storage.parent.scene().removeItem(selectedDirectPortPair.toPort)

    def removeHxL(self):
        self._removeSelectedHeatExchangers(self.leftHeatExchangersItemListWidget)

    def removeHxR(self):
        self._removeSelectedHeatExchangers(self.rightHeatExchangersItemListWidget)

    def _removeSelectedHeatExchangers(self, heatExchangersItemListWidget):
        for selectedItem in heatExchangersItemListWidget.selectedItems():
            heatExchanger = selectedItem.data(_qtc.Qt.UserRole)

            self.storage.heatExchangers.remove(heatExchanger)

            row = heatExchangersItemListWidget.row(selectedItem)
            heatExchangersItemListWidget.takeItem(row)

            while len(heatExchanger.port1.connectionList) > 0:
                heatExchanger.port1.connectionList[0].deleteConn()

            while len(heatExchanger.port2.connectionList) > 0:
                heatExchanger.port2.connectionList[0].deleteConn()

            self.storage.inputs.remove(heatExchanger.port1)
            self.storage.outputs.remove(heatExchanger.port2)

            self.storage.parent.scene().removeItem(heatExchanger.port1)
            self.storage.parent.scene().removeItem(heatExchanger.port2)
            self.storage.parent.scene().removeItem(heatExchanger)

    def modifyHx(self):
        """
        Modify Hx.
        """
        result = self._getFirstSelectedItemAndHeatExchanger()
        if not result:
            return
        selectedItem, heatExchanger = result

        modifyDialog = _mhd.ModifyRelativeHeightsDialog(
            heatExchanger.relativeInputHeight, heatExchanger.relativeOutputHeight
        )
        newHeights = modifyDialog.newRelativeHeights
        if not newHeights:
            return

        newInputHeight = newHeights.input if newHeights.input != "empty" else heatExchanger.relativeInputHeight
        newOutputHeight = newHeights.output if newHeights.output != "empty" else heatExchanger.relativeOutputHeight
        heatExchanger.setRelativeHeights(newInputHeight, newOutputHeight)

        listText = self._getHeatExchangerListItemText(heatExchanger)
        selectedItem.setText(listText)

    def _getFirstSelectedItemAndHeatExchanger(self):
        leftSelectedItems = self.leftHeatExchangersItemListWidget.selectedItems()
        rightSelectedItems = self.rightHeatExchangersItemListWidget.selectedItems()

        if leftSelectedItems:
            side = 0
            selectedItem = leftSelectedItems[0]
        elif rightSelectedItems:
            side = 2
            selectedItem = rightSelectedItems[0]
        else:
            return None

        name = selectedItem.text().split(",")[0]
        for heatExchanger in self.storage.heatExchangers:
            if heatExchanger.displayName == name and heatExchanger.sSide == side:
                return selectedItem, heatExchanger

        raise AssertionError(f"No heat exchanger with name {name} found.")

    def modifyPort(self):
        """
        Modify existing ports.
        """
        selectedItem = self._getFirstSelectedDirectPortPairListWidgetItem()
        if not selectedItem:
            return

        directPortPair: _dpp.DirectPortPair = selectedItem.data(_qtc.Qt.UserRole)

        dialogResult = _mhd.ModifyRelativeHeightsDialog(
            directPortPair.relativeInputHeight, directPortPair.relativeOutputHeight
        )
        newHeights = dialogResult.newRelativeHeights
        if not newHeights:
            return

        newRelativeInputHeight = newHeights.input if newHeights.input != "empty" else directPortPair.relativeInputHeight
        newRelativeOutputHeight = (
            newHeights.output if newHeights.output != "empty" else directPortPair.relativeOutputHeight
        )

        directPortPair.setRelativeHeights(newRelativeInputHeight, newRelativeOutputHeight, self.storage.h)

        newText = self._getDirectPortPairListItemText(directPortPair)
        selectedItem.setText(newText)

    def _getFirstSelectedDirectPortPairListWidgetItem(
            self,
    ) -> _tp.Optional[QListWidgetItem]:
        leftSelectedItems = self._leftDirectPortPairsItemListWidget.selectedItems()
        if leftSelectedItems:
            return leftSelectedItems[0]

        rightSelectedItems = self._rightDirectPortPairsItemListWidget.selectedItems()
        if rightSelectedItems:
            return rightSelectedItems[0]

        return None

    def incrSize(self):
        self._changeSize(self.HEIGHT_INCREMENT, self.WIDTH_INCREMENT)

    def decrSize(self):
        self._changeSize(-self.HEIGHT_INCREMENT, -self.WIDTH_INCREMENT)

    def _changeSize(self, deltaH, deltaW):
        self.storage.updatePortItemPositions(deltaH, deltaW)
        self.storage.h += deltaH
        self.storage.w += deltaW
        self.storage.updateHeatExchangersAfterTankSizeChange()
        self.storage.updateImage()

    def acceptedEdit(self):
        newName = self.lineEdit.text()
        if newName.lower() == str(self.storage.displayName).lower():
            self.close()
            return
        if self.lineEdit.text() == "":
            qmb = QMessageBox()
            qmb.setText("Please set a name for this storage tank.")
            qmb.setStandardButtons(QMessageBox.Ok)
            qmb.setDefaultButton(QMessageBox.Ok)
            qmb.exec()
            return
        if self.parent().nameExists(newName) or self.parent().nameExistsInDdckFolder(newName):
            msgb = QMessageBox()
            msgb.setText("Name already exist!")
            msgb.exec()
            return
        self.storage.setDisplayName(self.lineEdit.text())
        self.close()

    def cancel(self):
        self.close()



from __future__ import annotations

import typing as _tp

import PyQt5.QtWidgets as _qtw

import trnsysGUI.connection.connectionBase as _cb
import trnsysGUI.connection.deleteSinglePipeConnectionCommand as _dspc
import trnsysGUI.connection.singlePipeConnectionModel as _model
import trnsysGUI.connection.values as _values
import trnsysGUI.hydraulicLoops.names as _names
import trnsysGUI.massFlowSolver as _mfs
import trnsysGUI.massFlowSolver.networkModel as _mfn
import trnsysGUI.singlePipePortItem as _sppi
import trnsysGUI.singlePipeSegmentItem as _spsi
from trnsysGUI import PortItemBase as _pib, BlockItem as _bi, TVentil as _tventil

if _tp.TYPE_CHECKING:
    import trnsysGUI.diagram.Editor as _ed


class SinglePipeConnection(_cb.ConnectionBase):  # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        fromPort: _sppi.SinglePipePortItem,
        toPort: _sppi.SinglePipePortItem,
        parent: _ed.Editor  # type: ignore[name-defined]
    ):
        super().__init__(fromPort, toPort, parent)

        self._editor = parent

        self.diameterInCm: _values.Value = _values.DEFAULT_DIAMETER_IN_CM
        self.uValueInWPerM2K: _values.Value = _values.DEFAULT_U_VALUE_IN_W_PER_M2_K
        self.lengthInM: _values.Value = _values.DEFAULT_LENGTH_IN_M

    @property
    def fromPort(self) -> _sppi.SinglePipePortItem:
        assert isinstance(self._fromPort, _sppi.SinglePipePortItem)
        return self._fromPort

    @property
    def toPort(self) -> _sppi.SinglePipePortItem:
        assert isinstance(self._toPort, _sppi.SinglePipePortItem)
        return self._toPort

    def _createSegmentItem(self, startNode, endNode):
        return _spsi.SinglePipeSegmentItem(startNode, endNode, self)

    def getRadius(self):
        rad = 2
        return rad

    def editHydraulicLoop(self) -> None:
        self._editor.editHydraulicLoop(self)

    def createDeleteUndoCommand(self, parentCommand: _tp.Optional[_qtw.QUndoCommand] = None) -> _qtw.QUndoCommand:
        undoCommand = _dspc.DeleteSinglePipeConnectionCommand(
            self, self._editor.hydraulicLoops, self._editor.fluids.fluids, self._editor.fluids.WATER, parentCommand
        )

        return undoCommand

    def encode(self):
        if len(self.segments) > 0:
            labelPos = self.segments[0].label.pos().x(), self.segments[0].label.pos().y()
            labelMassPos = self.segments[0].labelMass.pos().x(), self.segments[0].labelMass.pos().y()
        else:
            self.logger.debug("This connection has no segment")
            defaultPos = self.fromPort.pos().x(), self.fromPort.pos().y()  # pylint: disable = duplicate-code # 2
            labelPos = defaultPos
            labelMassPos = defaultPos

        corners = []
        for s in self.getCorners():  # pylint: disable=invalid-name
            cornerTupel = (s.pos().x(), s.pos().y())
            corners.append(cornerTupel)

        connectionModel = _model.ConnectionModel(
            self.connId,
            self.displayName,
            self.id,
            corners,
            labelPos,
            labelMassPos,
            self.fromPort.id,
            self.toPort.id,
            self.trnsysId,
            self.diameterInCm,
            self.uValueInWPerM2K,
            self.lengthInM,
        )

        dictName = "Connection-"
        return dictName, connectionModel.to_dict()

    def decode(self, i):
        model = _model.ConnectionModel.from_dict(i)

        self.id = model.id
        self.connId = model.connectionId
        self.trnsysId = model.trnsysId
        self.setDisplayName(model.name)

        if len(model.segmentsCorners) > 0:
            self.loadSegments(model.segmentsCorners)

        self.setLabelPos(model.labelPos)
        self.setMassLabelPos(model.massFlowLabelPos)

        self.diameterInCm = model.diameterInCm
        self.uValueInWPerM2K = model.uValueInWPerM2K
        self.lengthInM = model.lengthInM

    def getInternalPiping(self) -> _mfs.InternalPiping:
        fromPort = _mfn.PortItem("Input", _mfn.PortItemType.INPUT)
        toPort = _mfn.PortItem("Output", _mfn.PortItemType.OUTPUT)

        pipe = _mfn.Pipe(self.displayName, self.trnsysId, fromPort, toPort)
        return _mfs.InternalPiping([pipe], {fromPort: self.fromPort, toPort: self.toPort})

    def exportPipeAndTeeTypesForTemp(self, startingUnit):  # pylint: disable=too-many-locals, too-many-statements
        lines = ""
        unitNumber = startingUnit
        typeNr2 = 931  # Temperature calculation from a pipe

        unitText = ""
        ambientT = 20

        loop = self._editor.hydraulicLoops.getLoopForExistingConnection(self)

        densityVar = _names.getDensityName(loop.name.value)
        specHeatVar = _names.getHeatCapacityName(loop.name.value)

        equationConstant1 = 1
        equationConstant2 = 3

        parameterNumber = 6
        inputNumbers = 4

        # Fixed strings
        tempRoomVar = "TRoomStore"
        initialValueS = "20 0.0 20 20"
        powerPrefix = "P"

        # Momentarily hardcoded
        equationNr = 3

        unitText += "UNIT " + str(unitNumber) + " TYPE " + str(typeNr2) + "\n"
        unitText += "!" + self.displayName + "\n"
        unitText += "PARAMETERS " + str(parameterNumber) + "\n"

        lengthInM = _getConvertedValueOrName(self.lengthInM)
        diameterInM = _getConvertedValueOrName(self.diameterInCm, 1 / 100)
        uValueInkJPerHourM2K = _getConvertedValueOrName(self.uValueInWPerM2K, 60 * 60 / 1000)

        unitText += f"{diameterInM} ! diameter [m]\n"
        unitText += f"{lengthInM} ! length [m]\n"

        uValueInSIUnitsComment = (
            f" (= {self.uValueInWPerM2K} W/(m^2*K))" if isinstance(self.uValueInWPerM2K, float) else ""
        )
        unitText += f"{uValueInkJPerHourM2K} ! U-value [kJ/(h*m^2*K)]{uValueInSIUnitsComment}\n"

        unitText += densityVar + "\n"
        unitText += specHeatVar + "\n"
        unitText += str(ambientT) + "\n"

        unitText += "INPUTS " + str(inputNumbers) + "\n"

        openLoops, _ = self._getOpenLoopsAndNodeToIndices()
        assert len(openLoops) == 1
        openLoop = openLoops[0]

        assert len(openLoop.realNodes) == 1
        realNode = openLoop.realNodes[0]

        outputVariables = realNode.getOutputVariables()

        portItemsWithParent = self._getFromAndToPortsAndParentBlockItems()

        if len(portItemsWithParent) == 2:
            portItem = portItemsWithParent[0][0]
            parent = portItemsWithParent[0][1]
            if hasattr(parent, "getSubBlockOffset"):
                unitText += "T" + parent.displayName + "X" + str(parent.getSubBlockOffset(self) + 1) + "\n"
            else:
                unitText += parent.getTemperatureVariableName(portItem) + "\n"

            unitText += f"{outputVariables[0].name}\n"
            unitText += tempRoomVar + "\n"

            portItem = portItemsWithParent[1][0]
            parent = portItemsWithParent[1][1]
            if hasattr(parent, "getSubBlockOffset"):
                unitText += "T" + parent.displayName + "X" + str(parent.getSubBlockOffset(self) + 1) + "\n"
            else:
                unitText += parent.getTemperatureVariableName(portItem) + "\n"

        else:
            lines += (
                "Error: NO VALUE\n" * 3
                + "at connection with parents "
                + str(self.fromPort.parent)
                + str(self.toPort.parent)
                + "\n"
            )

        unitText += "***Initial values\n"
        unitText += initialValueS + "\n\n"

        unitText += "EQUATIONS " + str(equationNr) + "\n"
        unitText += "T" + self.displayName + "= [" + str(unitNumber) + "," + str(equationConstant1) + "]\n"
        unitText += (
            powerPrefix
            + self.displayName
            + "_kW"
            + "= ["
            + str(unitNumber)
            + ","
            + str(equationConstant2)
            + "]/3600 !kW\n"
        )
        unitText += "Mfr" + self.displayName + "= " + "Mfr" + self.displayName + "_A\n\n"

        unitNumber += 1

        return unitText, unitNumber

    def _getFromAndToPortsAndParentBlockItems(
        self,
    ) -> _tp.Tuple[_tp.Tuple[_pib.PortItemBase, _bi.BlockItem], _tp.Tuple[_pib.PortItemBase, _bi.BlockItem]]:
        isToPortValveOutput = (
            isinstance(self.toPort.parent, _tventil.TVentil) and self.fromPort in self.toPort.parent.outputs
        )
        if isToPortValveOutput:
            return (self.toPort, self.toPort.parent), (self.fromPort, self.fromPort.parent)

        return (self.fromPort, self.fromPort.parent), (self.toPort, self.toPort.parent)


def _getConvertedValueOrName(valueOrName: _values.Value, conversionFactor=1.0) -> _tp.Union[float, str]:
    if isinstance(valueOrName, _values.Variable):
        return valueOrName.name

    value = valueOrName

    return value * conversionFactor



# pylint: skip-file

import dataclasses as _dc
import typing as _tp
import uuid as _uuid

import dataclasses_jsonschema as _dcj

import trnsysGUI.images as _img
import trnsysGUI.massFlowSolver.networkModel as _mfn
import pytrnsys.utils.serialization as _ser
from trnsysGUI.BlockItem import BlockItem  # type: ignore[attr-defined]
from trnsysGUI.connectors.doublePipeConnectorBase import DoublePipeBlockItemModel
from trnsysGUI.doublePipeModelPortItems import ColdPortItem, HotPortItem
from trnsysGUI.doublePipePortItem import DoublePipePortItem  # type: ignore[attr-defined]
from trnsysGUI.massFlowSolver import InternalPiping, MassFlowNetworkContributorMixin


class DoublePipeTeePiece(BlockItem, MassFlowNetworkContributorMixin):
    def __init__(self, trnsysType, parent, **kwargs):
        super().__init__(trnsysType, parent, **kwargs)

        self.w = 60
        self.h = 40

        self.inputs.append(DoublePipePortItem("i", 0, self))
        self.outputs.append(DoublePipePortItem("o", 2, self))
        self.outputs.append(DoublePipePortItem("o", 2, self))

        self.changeSize()

        self.childIds = []
        self.childIds.append(self.trnsysId)
        self.childIds.append(self.parent.parent().idGen.getTrnsysID())

    def _getImageAccessor(self) -> _tp.Optional[_img.ImageAccessor]:
        rotationAngle = (self.rotationN % 4) * 90

        if rotationAngle == 0:
            return _img.DP_TEE_PIECE_SVG

        if rotationAngle == 90:
            return _img.DP_TEE_PIECE_ROTATED_90

        if rotationAngle == 180:
            return _img.DP_TEE_PIECE_ROTATED_180

        if rotationAngle == 270:
            return _img.DP_TEE_PIECE_ROTATED_270

        raise AssertionError("Invalid rotation angle.")

    def changeSize(self):
        width, _ = self._getCappedWithAndHeight()
        self._positionLabel()

        self.origInputsPos = [[0, 30]]
        self.origOutputsPos = [[width, 30], [30, 0]]

        self.inputs[0].setPos(self.origInputsPos[0][0], self.origInputsPos[0][1])
        self.outputs[0].setPos(self.origOutputsPos[0][0], self.origOutputsPos[0][1])
        self.outputs[1].setPos(self.origOutputsPos[1][0], self.origOutputsPos[1][1])

        # pylint: disable=duplicate-code  # 1
        self.updateFlipStateH(self.flippedH)
        self.updateFlipStateV(self.flippedV)

        self.inputs[0].side = (self.rotationN + 2 * self.flippedH) % 4
        self.outputs[0].side = (self.rotationN + 2 - 2 * self.flippedH) % 4
        # pylint: disable=duplicate-code  # 1
        self.outputs[1].side = (self.rotationN + 1 - 2 * self.flippedV) % 4

    def encode(self):
        portListInputs = []
        portListOutputs = []

        for inp in self.inputs:
            portListInputs.append(inp.id)
        for output in self.outputs:
            portListOutputs.append(output.id)

        blockPosition = (float(self.pos().x()), float(self.pos().y()))

        doublePipeTeePieceModel = DoublePipeTeePieceModel(
            self.name,
            self.displayName,
            blockPosition,
            self.id,
            self.trnsysId,
            self.childIds,
            portListInputs,
            portListOutputs,
            self.flippedH,
            self.flippedV,
            self.rotationN,
        )

        dictName = "Block-"
        return dictName, doublePipeTeePieceModel.to_dict()

    def decode(self, i, resBlockList):
        model = DoublePipeTeePieceModel.from_dict(i)

        self.setDisplayName(model.BlockDisplayName)
        self.setPos(float(model.blockPosition[0]), float(model.blockPosition[1]))
        self.id = model.id
        self.trnsysId = model.trnsysId
        self.childIds = model.childIds

        self.inputs[0].id = model.portsIdsIn[0]
        self.outputs[0].id = model.portsIdsOut[0]
        self.outputs[1].id = model.portsIdsOut[1]

        self.updateFlipStateH(model.flippedH)
        self.updateFlipStateV(model.flippedV)
        self.rotateBlockToN(model.rotationN)

        resBlockList.append(self)


    def getInternalPiping(self) -> InternalPiping:
        coldInput: _mfn.PortItem = ColdPortItem("coldInput", _mfn.PortItemType.INPUT)
        coldOutput1: _mfn.PortItem = ColdPortItem("coldStraightOutput", _mfn.PortItemType.OUTPUT)
        coldOutput2: _mfn.PortItem = ColdPortItem("coldOrthogonalOutput", _mfn.PortItemType.OUTPUT)
        coldTeePiece = _mfn.TeePiece(self.displayName + "Cold", self.childIds[0], coldInput, coldOutput1, coldOutput2)
        coldModelPortItemsToGraphicalPortItem = {
            coldInput: self.inputs[0], coldOutput1: self.outputs[0], coldOutput2: self.outputs[1]}

        hotInput: _mfn.PortItem = HotPortItem("hotInput", _mfn.PortItemType.INPUT)
        hotOutput1: _mfn.PortItem = HotPortItem("hotStraightOutput", _mfn.PortItemType.OUTPUT)
        hotOutput2: _mfn.PortItem = HotPortItem("hotOrthogonalOutput", _mfn.PortItemType.OUTPUT)
        hotTeePiece = _mfn.TeePiece(self.displayName + "Hot", self.childIds[1], hotInput, hotOutput1, hotOutput2)
        hotModelPortItemsToGraphicalPortItem = {
            hotInput: self.inputs[0], hotOutput1: self.outputs[0], hotOutput2: self.outputs[1]}

        modelPortItemsToGraphicalPortItem = coldModelPortItemsToGraphicalPortItem | hotModelPortItemsToGraphicalPortItem

        internalPiping = InternalPiping([coldTeePiece, hotTeePiece], modelPortItemsToGraphicalPortItem)

        return internalPiping

    def exportPipeAndTeeTypesForTemp(self, startingUnit): # pylint: disable=too-many-locals
        unitNumber = startingUnit
        tNr = 929  # Temperature calculation from a tee-piece

        unitText = ""
        ambientT = 20
        equationConstant = 1

        openLoops, nodesToIndices = self._getOpenLoopsAndNodeToIndices()
        assert len(openLoops) == 2
        coldOpenLoop = openLoops[0]
        hotOpenLoop = openLoops[1]

        unitNumber, unitText = self._getExport(ambientT, equationConstant, nodesToIndices, coldOpenLoop, tNr,
                                                "Cold", unitNumber, unitText)
        unitNumber, unitText = self._getExport(ambientT, equationConstant, nodesToIndices, hotOpenLoop, tNr, "Hot",
                                                unitNumber, unitText)
        return unitText, unitNumber

    def _getExport(self, ambientT, equationConstant, nodesToIndices, openLoop, tNr, temperature, unitNumber, unitText):
        unitText += "UNIT " + str(unitNumber) + " TYPE " + str(tNr) + "\n"
        unitText += "!" + self.displayName + temperature + "\n"
        unitText += "PARAMETERS 0\n"
        unitText += "INPUTS 6\n"

        assert len(openLoop.realNodes) == 1
        realNode = openLoop.realNodes[0]

        outputVariables = realNode.getOutputVariables()
        for outputVariable in outputVariables:
            if not outputVariable:
                continue

            unitText += outputVariable.name + "\n"

        unitText += f"T{self.inputs[0].connectionList[0].displayName}{temperature}\n"
        unitText += f"T{self.outputs[0].connectionList[0].displayName}{temperature}\n"
        unitText += f"T{self.outputs[1].connectionList[0].displayName}{temperature}\n"

        unitText += "***Initial values\n"
        unitText += 3 * "0 " + 3 * (str(ambientT) + " ") + "\n"

        unitText += "EQUATIONS 1\n"
        unitText += "T" + self.displayName + temperature + "= [" + str(unitNumber) + "," + str(equationConstant) + "]\n\n"
        unitNumber += 1

        return unitNumber, unitText

@_dc.dataclass
class DoublePipeTeePieceModel(
    _ser.UpgradableJsonSchemaMixin):  # pylint: disable=too-many-instance-attributes
    BlockName: str  # pylint: disable=invalid-name
    BlockDisplayName: str  # pylint: disable=invalid-name
    blockPosition: _tp.Tuple[float, float]
    id: int  # pylint: disable=invalid-name
    trnsysId: int
    childIds: _tp.List[int]
    portsIdsIn: _tp.List[int]
    portsIdsOut: _tp.List[int]
    flippedH: bool
    flippedV: bool
    rotationN: int

    @classmethod
    def from_dict(
        cls,
        data: _dcj.JsonDict,
        validate=True,  # pylint: disable=duplicate-code
        validate_enums: bool = True,
    ) -> "DoublePipeTeePieceModel":
        doublePipeTeePieceModel = super().from_dict(data, validate, validate_enums)
        return _tp.cast(DoublePipeTeePieceModel, doublePipeTeePieceModel)

    def to_dict(
        self,
        omit_none: bool = True,
        validate: bool = False,
        validate_enums: bool = True,
    ) -> _dcj.JsonDict:
        data = super().to_dict(omit_none, validate, validate_enums)  # pylint: disable=duplicate-code
        data[".__BlockDict__"] = True
        return data

    @classmethod
    def getSupersededClass(cls) -> _tp.Type[_ser.UpgradableJsonSchemaMixinVersion0]:
        return DoublePipeBlockItemModel

    @classmethod
    def upgrade(cls, superseded: DoublePipeBlockItemModel) -> "DoublePipeTeePieceModel": # type: ignore[override]
        assert len(superseded.portsIdsIn) == 2
        assert len(superseded.portsIdsOut) == 1

        inputPortIds = [superseded.portsIdsIn[0]]
        outputPortIds = [superseded.portsIdsIn[1], superseded.portsIdsOut[0]]

        return DoublePipeTeePieceModel(
            superseded.BlockName,
            superseded.BlockDisplayName,
            superseded.blockPosition,
            superseded.id,
            superseded.trnsysId,
            superseded.childIds,
            inputPortIds,
            outputPortIds,
            superseded.flippedH,
            superseded.flippedV,
            superseded.rotationN,
        )

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID('3fff9a8a-d40e-42e2-824d-c015116d0a1d')


from __future__ import annotations

import typing as _tp

import trnsysGUI.PortItemBase as _pi
import trnsysGUI.massFlowSolver as _mfs
import trnsysGUI.massFlowSolver.networkModel as _mfn

if _tp.TYPE_CHECKING:
    import trnsysGUI.BlockItem as _bi

GetInternallyConnectedPortItems = _tp.Callable[["SinglePipePortItem"], _tp.Sequence["SinglePipePortItem"]]
Side = _tp.Literal[0, 2]


class SinglePipePortItem(_pi.PortItemBase):
    def __init__(
            self,
            name: str,
            side: Side,
            parent: _bi.BlockItem,
            getInternallyConnectedPortItems: GetInternallyConnectedPortItems,
    ) -> None:
        super().__init__(name, side, parent)

        self._getInternallyConnectedPortItems = getInternallyConnectedPortItems

    def _selectConnectedRealNode(  # pylint: disable=duplicate-code  # 2
            self,
            portItem: _mfn.PortItem,
            connectedPortItemsAndAdjacentRealNode: _tp.Sequence[_mfs.PortItemAndAdjacentRealNode],
    ) -> _mfn.RealNodeBase:
        assert (
                len(connectedPortItemsAndAdjacentRealNode) == 1
        ), "Only exactly one model port item should be connected with a single pipe graphical port item."

        selectedRealNode = connectedPortItemsAndAdjacentRealNode[0].realNode

        return selectedRealNode

    def _highlightInternallyConnectedPortItems(self):
        for portItem in self._getInternallyConnectedPortItems(self):
            portItem.innerCircle.setBrush(self.ashColorB)

    def _unhighlightInternallyConnectedPortItems(self):
        for portItem in self._getInternallyConnectedPortItems(self):
            portItem.innerCircle.setBrush(self.visibleColor)



from __future__ import annotations

import trnsysGUI.connection.values as _values
from . import _loopWideDefaults as _lwd
from . import common as _common
from . import model as _model
from . import names as _names
from ._dialogs.edit import dialog as _gui, model as _gmodel


def edit(hydraulicLoop: _model.HydraulicLoop, hydraulicLoops: _model.HydraulicLoops, fluids: _model.Fluids) -> None:
    guiHydraulicLoop = _createGuiLoop(hydraulicLoop)

    occupiedNames = [l.name.value for l in hydraulicLoops.hydraulicLoops if l != hydraulicLoop]

    _common.setConnectionsSelected(hydraulicLoop.connections, True)

    okedOrCancelled = _gui.HydraulicLoopDialog.showDialog(guiHydraulicLoop, occupiedNames, fluids)

    _common.setConnectionsSelected(hydraulicLoop.connections, False)

    if okedOrCancelled == "cancelled":
        return

    _applyModel(guiHydraulicLoop, hydraulicLoop)


def _createGuiLoop(hydraulicLoop: _model.HydraulicLoop) -> _gmodel.HydraulicLoop:
    guiConnections = _createGuiConnections(hydraulicLoop)

    guiLoop = _gmodel.HydraulicLoop(
        hydraulicLoop.name.value, hydraulicLoop.fluid, hydraulicLoop.useLoopWideDefaults, guiConnections
    )
    return guiLoop


def _createGuiConnections(hydraulicLoop):
    if hydraulicLoop.useLoopWideDefaults:
        guiConnections = [
            _gmodel.Connection(
                c.displayName,
                _values.DEFAULT_DIAMETER_IN_CM,
                _values.DEFAULT_U_VALUE_IN_W_PER_M2_K,
                _values.DEFAULT_LENGTH_IN_M,
                c,
            )
            for c in hydraulicLoop.connections
        ]

        return guiConnections

    guiConnections = [
        _gmodel.Connection(c.displayName, c.diameterInCm, c.uValueInWPerM2K, c.lengthInM, c)
        for c in hydraulicLoop.connections
    ]

    return guiConnections


def _applyModel(guiHydraulicLoop: _gmodel.HydraulicLoop, hydraulicLoop: _model.HydraulicLoop) -> None:
    oldName = hydraulicLoop.name.value
    newName = guiHydraulicLoop.name
    if oldName != newName:
        hydraulicLoop.name = _model.UserDefinedName(newName)

    hydraulicLoop.fluid = guiHydraulicLoop.fluid
    hydraulicLoop.useLoopWideDefaults = guiHydraulicLoop.useLoopWideDefaults

    if guiHydraulicLoop.useLoopWideDefaults:
        _lwd.resetConnectionPropertiesToLoopWideDefaults(hydraulicLoop.connections, hydraulicLoop.name.value)
    else:
        _applyConnectionModels(guiHydraulicLoop)


def _applyConnectionModels(guiHydraulicLoop: _gmodel.HydraulicLoop) -> None:
    loopName = guiHydraulicLoop.name
    useLoopWideDefaults = guiHydraulicLoop.useLoopWideDefaults

    for modelConnection in guiHydraulicLoop.connections:
        connection = modelConnection.connection

        if useLoopWideDefaults:
            connection.lengthInM = _values.Variable(_names.getDefaultLengthName(loopName))
            connection.diameterInCm = _values.Variable(_names.getDefaultDiameterName(loopName))
            connection.uValueInWPerM2K = _values.Variable(_names.getDefaultUValueName(loopName))
        else:
            connection.lengthInM = modelConnection.lengthInM
            connection.diameterInCm = modelConnection.diameterInCm
            connection.uValueInWPerM2K = modelConnection.uValueInWPerM2K



import typing as _tp

import PyQt5.QtWidgets as _qtw
import PyQt5.QtGui as _qtg
import PyQt5.QtCore as _qtc

from trnsysGUI.hydraulicLoops import common as _common

try:
    from . import _UI_dialog_generated as _uigen
except ImportError as importError:
    raise AssertionError(  # pylint: disable=duplicate-code  # 2
        "Could not find the generated Python code for a .ui or .qrc file. Please run the "
        "`dev-tools\\generateGuiClassesFromQtCreatorStudioUiFiles.py' script from your "
        "`pytrnsys_gui` directory."
    ) from importError

from trnsysGUI.hydraulicLoops import model as _model


class SplitLoopDialog(_qtw.QDialog, _uigen.Ui_splitHydraulicLoopDialog):
    @staticmethod
    def showDialogAndGetResult(
        loop: _model.HydraulicLoop,
        occupiedNames: _tp.Set[str],
        fluids: _tp.Sequence[_model.Fluid],
        setLoop1Selected: _tp.Callable[[bool], None],
        setLoop2Selected: _tp.Callable[[bool], None]
    ) -> _common.Cancellable[_common.SplitLoopsSummary]:
        dialog = SplitLoopDialog(loop, occupiedNames, fluids, setLoop1Selected, setLoop2Selected)
        dialog.exec()
        return dialog.splitSummary

    def __init__(
        self,
        loop: _model.HydraulicLoop,
        occupiedNames: _tp.Set[str],
        fluids: _tp.Sequence[_model.Fluid],
        setLoop1Selected: _tp.Callable[[bool], None],
        setLoop2Selected: _tp.Callable[[bool], None]
    ) -> None:
        super().__init__()
        self.setupUi(self)

        self._loop = loop
        self._occupiedNames = occupiedNames
        self._fluids = fluids
        self._setLoop1Selected = setLoop1Selected
        self._setLoop2Selected = setLoop2Selected

        self.splitSummary: _common.Cancellable[_common.SplitLoopsSummary] = "cancelled"

        self._configureGroupBoxes()

        self._configureNameComboBoxes()
        self.name1ComboBox.setCurrentIndex(0)
        self.name2ComboBox.setCurrentIndex(0)

        self._configureFluidComboBoxes()
        self.fluid1ComboBox.setCurrentIndex(0)
        self.fluid2ComboBox.setCurrentIndex(0)

        self._configureButtons()

        self._onAnyChange(self.name1ComboBox.currentText(), self.name2ComboBox.currentText())

    def _configureGroupBoxes(self) -> None:
        def onEnter1() -> None:
            self._setLoop1Selected(True)

        self.loop1GroupBox.enter.connect(onEnter1)

        def onLeave1() -> None:
            self._setLoop1Selected(False)

        self.loop1GroupBox.leave.connect(onLeave1)

        def onEnter2() -> None:
            self._setLoop2Selected(True)

        self.loop2GroupBox.enter.connect(onEnter2)

        def onLeave2() -> None:
            self._setLoop2Selected(False)

        self.loop2GroupBox.leave.connect(onLeave2)

    def _configureButtons(self) -> None:
        def deselectAllConnections() -> None:
            self._setLoop1Selected(False)
            self._setLoop2Selected(False)

        def onAbort() -> None:
            self.splitSummary = "cancelled"
            self.close()

        self.abortButton.clicked.connect(deselectAllConnections)
        self.abortButton.clicked.connect(onAbort)

        def onApply() -> None:
            name1Value = self._createNameForValue(self.name1ComboBox.currentText())
            name1 = name1Value
            name2Value = self.name2ComboBox.currentText()
            name2 = self._createNameForValue(name2Value)

            fluid1 = self.fluid1ComboBox.currentData()
            fluid2 = self.fluid2ComboBox.currentData()

            self.splitSummary = _common.SplitLoopsSummary(
                _common.LoopSummary(name1, fluid1), _common.LoopSummary(name2, fluid2)
            )

            self.close()

        self.applyButton.clicked.connect(deselectAllConnections)
        self.applyButton.clicked.connect(onApply)

    def _createNameForValue(self, nameValue: str) -> _model.Name:
        if nameValue != self._loop.name.value:
            return _model.UserDefinedName(nameValue)

        if self._loop.name.isUserDefined:
            return _model.UserDefinedName(nameValue)

        return _model.AutomaticallyGeneratedName(nameValue)

    def _configureFluidComboBoxes(self) -> None:
        self._configureFluidComboBox(self.fluid1ComboBox)
        self._configureFluidComboBox(self.fluid2ComboBox)

    def _configureFluidComboBox(self, fluidComboBox: _qtw.QComboBox) -> None:
        fluidComboBox.addItem(self._loop.fluid.name, self._loop.fluid)
        boldFont = _qtg.QFont()
        boldFont.setBold(True)
        fluidComboBox.setItemData(0, boldFont, _qtc.Qt.FontRole)
        otherFluids = {*self._fluids} - {self._loop.fluid}
        sortedOtherFluids = sorted(otherFluids, key=lambda f: f.name)
        for otherFluid in sortedOtherFluids:
            fluidComboBox.addItem(otherFluid.name, otherFluid)

    def _configureNameComboBoxes(self) -> None:
        self.name1ComboBox.addItem("")
        self.name1ComboBox.addItem(self._loop.name.value)

        def onName1TextChanged(newName: str) -> None:
            self._onAnyChange(
                newName,
                self.name2ComboBox.currentText(),
            )

        self.name1ComboBox.currentTextChanged.connect(onName1TextChanged)

        self.name2ComboBox.addItem("")
        self.name2ComboBox.addItem(self._loop.name.value)

        def onName2TextChanged(newName: str) -> None:
            self._onAnyChange(
                self.name1ComboBox.currentText(),
                newName,
            )

        self.name2ComboBox.currentTextChanged.connect(onName2TextChanged)

    def _onAnyChange(self, name1: str, name2: str) -> None:
        areNamesOk = self._onEitherNameChanged(name1, name2)
        self.applyButton.setEnabled(areNamesOk)

    def _onEitherNameChanged(self, name1: str, name2: str) -> bool:
        isName1Valid = self._onNameChanged(name1, self.name1WarningWidget)
        isName2Valid = self._onNameChanged(name2, self.name2WarningWidget)

        if not isName1Valid or not isName2Valid:
            return False

        if name1 == name2:
            tooltip = "The names of the two loops must differ"
            self.name1WarningWidget.setToolTip(tooltip)
            self.name1WarningWidget.setVisible(True)
            self.name2WarningWidget.setToolTip(tooltip)
            self.name2WarningWidget.setVisible(True)

            return False

        return True

    def _onNameChanged(self, newName: str, warningWidget: _qtw.QWidget) -> bool:
        isEmpty = not newName
        isInUse = newName in self._occupiedNames
        tooltip = ""
        if isEmpty:
            tooltip = "You must specify a name"
        elif isInUse:
            tooltip = "This name is already in use"

        isValid = not isEmpty and not isInUse

        warningWidget.setVisible(not isValid)
        warningWidget.setToolTip(tooltip)

        return isValid
